import typing
from enum import StrEnum

from pydantic import BaseModel


class WorkflowState(StrEnum):
    CREATED = "CREATED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    EXPIRED = "EXPIRED"


class StepExecutionStatus(StrEnum):
    SUCCESS = "COMPLETED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    STARTED = "STARTED"
    UNKNOWN = "UNKNOWN"


class FlowMessage(BaseModel):
    """A message that can be sent between steps in a workflow.It's the only parameter step takes as input."""

    payload: typing.Any = None  # The payload of the message
    headers: dict[str, str] = None  # The headers of the message
    output_text: str = None  # The output text of the step that is captured in the execution log
    error_text: str = None  # The error text of the step that is captured in the execution log
    next_step_ids: list[str] = None  # If set, the workflow will skip default route and go to the next step in the list
    step_execution_status: StepExecutionStatus = StepExecutionStatus.UNKNOWN  # The status of the step execution


class StepType(StrEnum):
    PYSTEP = "pystep"
    START_WORKFLOW_TASK_STEP = "start_workflow_task_step"  # Call a workflow from another workflow
    CONTAINER_TASK_STEP = "container_step"  # TODO: implement
    CLI_TASK_STEP = "cli_step"  # TODO: implement
    HTTP_TRIGGER = "http_trigger"
    TIME_TRIGGER = "time_trigger"
    EVENT_TRIGGER = "event_trigger"  # TODO: implement


class UIConfig(BaseModel):
    pos_x: int = 0
    pos_y: int = 0


class WorkflowConfigItem(BaseModel):
    name: str
    value: str = None
    label: str = None
    type: str = None
    required: bool = False
    options: list[str] = None
    group: str = None


class WorkflowStepDefinition(BaseModel):
    id: str
    label: str = None
    stype: str = StepType.PYSTEP
    description: str = None
    method: str = None
    enabled: bool = True
    group_id: str = None
    trigger: bool = False
    transition_to: list[str] = None
    ui_config: UIConfig = UIConfig()
    params: dict[str, str] = None


class WorkflowStepsGroup(BaseModel):
    # Container for steps
    id: str
    label: str
    transition_to: list[str] = None
    description: str = None
    ui_config: UIConfig = UIConfig()


class WorkflowDefinition(BaseModel):
    """Workflow definition"""

    name: str
    description: str = None
    implementation_module: str = None
    steps: list[WorkflowStepDefinition]
    groups: list[WorkflowStepsGroup]
    configs: list[WorkflowConfigItem] = None


class WorkflowStepEvent(BaseModel):
    """Workflow step event represent a single step execution"""

    id: str
    group_id: str = None
    state: StepExecutionStatus
    elapsed_time: float
    timestamp: str
    error: str = None
    output_text: str = None
    data: typing.Any = None


class WorkflowFullStateReport(BaseModel):
    """Workflow state report is complete log of workflow execution"""

    workflow_name: str = None
    workflow_version: str = None
    run_id: str = None
    state: WorkflowState
    start_time: int = None
    end_time: int = None
    elapsed_time: float = 0
    last_error: str = None
    execution_log: list[WorkflowStepEvent]



import importlib
import inspect
import logging
import os
import sys
import traceback
from pathlib import Path

from cognite.client import CogniteClient

from cognite.neat.core.workflow import BaseWorkflow
from cognite.neat.core.workflow.base import WorkflowDefinition
from cognite.neat.core.workflow.tasks import WorkflowTaskBuilder


class WorkflowManager:
    """Workflow manager is responsible for loading, saving and managing workflows
    client: CogniteClient
    registry_storage_type: str = "file"
    workflows_storage_path: Path = Path("workflows")
    rules_storage_path: Path = Path("rules")
    data_set_id: int = None,
    """

    def __init__(
        self,
        client: CogniteClient = None,
        registry_storage_type: str = "file",
        workflows_storage_path: Path = None,
        rules_storage_path: Path = None,
        data_set_id: int = None,
    ):
        self.client = client
        self.data_set_id = data_set_id
        self.workflow_registry: dict[str, BaseWorkflow] = {}
        self.workflows_storage_type = registry_storage_type
        # todo use pathlib
        self.workflows_storage_path = workflows_storage_path if workflows_storage_path else Path("workflows")
        self.rules_storage_path = rules_storage_path if rules_storage_path else Path("rules")
        self.task_builder = WorkflowTaskBuilder(client, self)

    def update_cdf_client(self, client: CogniteClient):
        self.client = client
        self.task_builder = WorkflowTaskBuilder(client, self)
        self.workflow_registry = {}
        self.load_workflows_from_storage_v2()

    def get_list_of_workflows(self):
        return list(self.workflow_registry.keys())

    def get_workflow(self, name: str) -> BaseWorkflow:
        return self.workflow_registry[name]

    def start_workflow(self, name: str, sync=False, **kwargs):
        workflow = self.get_workflow(name)
        workflow.start(sync, kwargs=kwargs)
        return workflow

    def delete_workflow(self, name: str):
        del self.workflow_registry[name]
        return

    def update_workflow(self, name: str, workflow: WorkflowDefinition):
        self.workflow_registry[name].workflow_steps = workflow.steps
        self.workflow_registry[name].configs = workflow.configs
        self.workflow_registry[name].workflow_step_groups = workflow.groups
        return

    def save_workflow_to_storage(self, name: str, custom_implementation_module: str = None):
        """Save workflow from memory to storage"""
        if self.workflows_storage_type == "file":
            full_path = self.workflows_storage_path / name / "workflow.yaml"
            wf = self.workflow_registry[name]
            with open(full_path, "w") as f:
                f.write(
                    wf.serialize_workflow(
                        output_format="yaml", custom_implementation_module=custom_implementation_module
                    )
                )

    def load_workflows_from_storage_v2(self, dir_path: str = None):
        """Loads workflows from disk/storage into memory , initializes and register them in the workflow registry"""
        if dir_path:
            dir_path = Path(dir_path)
        else:
            dir_path = self.workflows_storage_path
        sys.path.append(str(dir_path))
        for wf_module_name in os.listdir(dir_path):
            wf_module_full_path = dir_path / wf_module_name
            if wf_module_full_path.is_dir():
                try:
                    logging.info(f"Loading workflow {wf_module_name} from {wf_module_full_path}")
                    # metadata_file = f"{dir_path}//{module_name}.yaml"
                    metadata_file = dir_path / wf_module_name / "workflow.yaml"
                    logging.info(f"Loading workflow {wf_module_name} metadata from {metadata_file}")
                    if os.path.exists(metadata_file):
                        with open(metadata_file, "r") as f:
                            wf_str = f.read()
                            metadata = BaseWorkflow.deserialize_metadata(wf_str, output_format="yaml")
                    else:
                        logging.info(f"Metadata file {metadata_file} not found, skipping")
                        continue

                    if metadata.implementation_module:
                        wf_module_name = metadata.implementation_module
                        logging.info(f"Loading CUSTOM workflow module {wf_module_name}")
                    else:
                        logging.info(f"Loading workflow module {wf_module_name}")

                    full_module_name = f"{wf_module_name}.workflow"
                    if full_module_name in sys.modules:
                        logging.info(f"Reloading existing workflow module {wf_module_name}")
                        module = importlib.reload(sys.modules[full_module_name])
                    else:
                        logging.info(f"Loading NEW workflow module {wf_module_name}")
                        module = importlib.import_module(full_module_name)
                    # dynamically load all classes in the module
                    for name, obj in inspect.getmembers(module):
                        if "NeatWorkflow" not in name:
                            continue
                        logging.info(f"Found class {name} in module {wf_module_name}")
                        if inspect.isclass(obj):
                            self.workflow_registry[wf_module_name] = obj(wf_module_name, self.client)
                            self.workflow_registry[wf_module_name].set_metadata(metadata)
                            self.workflow_registry[wf_module_name].set_task_builder(self.task_builder)
                            self.workflow_registry[wf_module_name].set_default_dataset_id(self.data_set_id)
                            self.workflow_registry[wf_module_name].set_storage_path(
                                "transformation_rules", self.rules_storage_path
                            )
                except Exception as e:
                    trace = traceback.format_exc()
                    logging.error(f"Error loading workflow {wf_module_name}: error: {e} trace : {trace}")



import logging

from cognite.client import CogniteClient

from cognite.neat.core.workflow.base import BaseWorkflow
from cognite.neat.core.workflow.model import FlowMessage


class BasicNeatWorkflow(BaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client, [])
        self.counter = 0

    def step_run_experiment_1(self, flow_msg: FlowMessage = None):
        logging.info("Running experiment 1")
        logging.info("Done running experiment 4444")
        self.counter = self.counter + 1
        logging.info("Counter: " + str(self.counter))
        if self.counter > 5:
            return FlowMessage(output_text="Done running experiment", next_step_ids=["error_handler"])
        else:
            return FlowMessage(
                output_text=f"Running iteration {self.counter} of xperiment", next_step_ids=["run_experiment_1"]
            )

    def step_cleanup(self, flow_msg: FlowMessage = None):
        logging.info("Cleanup")

    def step_error_handler(self, flow_msg: FlowMessage = None):
        logging.info("Error handler")
        return FlowMessage(output_text="Error handleed")



from pathlib import Path

import gspread
import pandas as pd
from openpyxl import Workbook, load_workbook

__all__ = [
    "google_to_table_by_name",
    "excel_file_to_table_by_name",
]


def google_to_table_by_name(sheet_id: str) -> dict[str, pd.DataFrame]:
    client_google = gspread.service_account()
    spreadsheet = client_google.open_by_key(sheet_id)
    return {worksheet.title: pd.DataFrame(worksheet.get_all_records()) for worksheet in spreadsheet.worksheets()}


def excel_file_to_table_by_name(filepath: Path) -> dict[str, pd.DataFrame]:
    workbook: Workbook = load_workbook(filepath)
    return {
        sheetname: pd.read_excel(
            filepath,
            sheet_name=sheetname,
            header=None if sheetname == "Metadata" else 0,
            skiprows=1 if sheetname in ["Classes", "Properties", "Instances"] else None,
        )
        for sheetname in workbook.sheetnames
    }



import logging
import warnings
from datetime import datetime, timezone
from typing import Dict, List, Optional, Tuple, Union
from warnings import warn

import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetHierarchy, AssetList
from deepdiff import DeepDiff
from rdflib import Graph, Namespace
from rdflib.term import URIRef

from cognite.neat.core.data_classes import AssetTemplate, Property
from cognite.neat.core.data_classes.config import EXCLUDE_PATHS
from cognite.neat.core.data_classes.transformation_rules import TransformationRules
from cognite.neat.core.loader.graph_store import NeatGraphStore
from cognite.neat.core.utils import chunker, datetime_utc_now, remove_namespace

ORPHANAGE = {
    "external_id": "orphanage",
    "name": "Orphanage",
    "parent_external_id": None,
    "description": "Used to store all assets which parent does not exist",
    "metadata": {
        "type": "Orphanage",
        "cdfResourceType": "Asset",
        "identifier": "orphanage",
        "active": "true",
        "start_time": "",
        "update_time": "",
    },
    "data_set_id": None,
    "labels": ["Orphanage", "non-historic"],
}


def _get_class_instance_ids(graph: Graph, class_: str, namespace: Namespace, limit: int = -1) -> List[URIRef]:
    """Get instances ids for a given class

    Parameters
    ----------
    graph : Graph
        Graph containing class instances
    class_ : str
        Class for which instances are to be found
    namespace : Namespace
        Namespace of given class (to avoid writing long URIs)
    limit : int
        Max number of instances to return, by default -1 meaning all instances

    Returns
    -------
    List[URIRef]
        List of class instance URIs
    """

    query_statement = "SELECT DISTINCT ?subject WHERE { ?subject a <class> .} LIMIT X".replace(
        "class", namespace[class_]
    ).replace("LIMIT X", "" if limit == -1 else f"LIMIT {limit}")
    logging.debug(f"Query statement: {query_statement}")
    return [res[0] for res in list(graph.query(query_statement))]


def _get_class_instance(graph: Graph, instance: URIRef) -> List[tuple]:
    """Get instance by means of tuples containing property-value pairs

    Returns
    -------
    List
        list property-value pairs
    """

    query_statement = "SELECT DISTINCT ?predicate ?object WHERE {<subject> ?predicate ?object .}".replace(
        "subject", instance
    )
    result = list(graph.query(query_statement))

    # Adds instance id for sake of keep chain of custody
    result += [(URIRef("http://purl.org/dc/terms/identifier"), instance)]

    return result


def _get_class_property_pairs(transformation_rules: TransformationRules) -> dict[List[Property]]:
    """Define classes in terms of their properties

    Parameters
    ----------
    transformation_rules : TransformationRules
        _description_

    Returns
    -------
    dict
        Dict containing keys as class names and list of their properties
    """

    classes = {}

    for property_ in transformation_rules.properties.keys():
        class_ = transformation_rules.properties[property_].class_name

        if class_ in classes:
            classes[class_] += [transformation_rules.properties[property_]]
        else:
            classes[class_] = [transformation_rules.properties[property_]]

    return classes


def _define_asset_class_mapping(transformation_rules: TransformationRules) -> dict:
    """Define mapping from class to asset properties

    Parameters
    ----------
    transformation_rules : TransformationRules
        Transformation rules

    Returns
    -------
    tuple[dict, dict[Any, dict[str, list[Any]]]]
        Dict containing mapping from class to asset properties
    """
    solution2cdf_mapping_rules = _get_class_property_pairs(transformation_rules)

    asset_class_mapping = {}

    for class_, properties in solution2cdf_mapping_rules.items():
        asset_class_mapping[class_] = {
            "external_id": [],
            "name": [],
            "description": [],
            "parent_external_id": [],
            "metadata": [],
        }

        for property_ in properties:
            if "Asset" in property_.cdf_resource_type and property_.property_name != "*":
                for resource_type_property in property_.resource_type_property:
                    if (
                        resource_type_property in asset_class_mapping[class_]
                        and property_.property_name not in asset_class_mapping[class_][resource_type_property]
                    ):
                        asset_class_mapping[class_][resource_type_property] += [property_.property_name]

                    if property_.property_name not in asset_class_mapping[class_]["metadata"]:
                        asset_class_mapping[class_]["metadata"] += [property_.property_name]

    return asset_class_mapping


def _remap_class_properties(class_instance: dict, asset_class_mapping: dict) -> Tuple[dict, set, set]:
    """Remaps original class instance properties to asset properties (e.g., external_id, name, description, metadata)

    Parameters
    ----------
    class_instance : dict
        Dictionary containing class instance properties and values originating from RDF
        stripped from namespaces
    asset_class_mapping : dict
        Property mapping from class to asset

    Returns
    -------
    Tuple[dict, set, set]
        Remapped class instance, set of missing asset properties and set of missing asset metadata
    """
    # Make distinction between missing properties that map into Asset fields
    # and missing RDF properties that are defined by sheet
    instance_properties = list(class_instance.keys())
    missing_properties = set()

    for property_group, ordered_properties in asset_class_mapping.items():
        if property_group != "metadata" and ordered_properties:
            if matching_property := next(
                (a for a in ordered_properties if a in instance_properties),
                None,
            ):
                class_instance[property_group] = class_instance[matching_property]
            else:
                missing_properties.add(property_group)

    missing_metadata = set(asset_class_mapping["metadata"]).difference(set(instance_properties))

    return class_instance, missing_properties, missing_metadata


def _class2asset_instance(
    class_: str,
    class_instance: dict,
    asset_class_mapping: dict,
    data_set_id: int,
    orphanage_asset_external_id: str = None,
    external_id_prefix: str = None,
    fallback_property: str = "identifier",
    empty_name_default: str = "Missing Name",
    add_missing_metadata: bool = True,
) -> dict:
    """Converts class instance to asset instance dictionary

    Parameters
    ----------
    class_ : str
        Class name which instance is being converted to asset instance
    class_instance : dict
        Dictionary containing class instance properties and values originating from RDF
        stripped from namespaces
    asset_class_mapping : dict
        Property mapping from class to asset
    data_set_id : int
        data set id to which asset belongs
    orphanage_asset_id : str, optional
        Orphanage asset external id, by default None
    external_id_prefix : str, optional
        External id prefix to be added to any external id, by default None
    fallback_property : str, optional
        Property from class instance to be used as fallback in case of missing properties, by default "identifier"


    Returns
    -------
    dict
        Asset instance dictionary
    """

    remapped_class_instance, missing_properties, missing_metadata = _remap_class_properties(
        class_instance, asset_class_mapping
    )

    # setting class instance type to class name
    remapped_class_instance["type"] = class_
    # This will be a default case since we want to use original identifier as external_id
    # We are though dropping namespace from the original identifier (avoiding long-tail URIs)
    if "external_id" in missing_properties or asset_class_mapping["external_id"] == []:
        try:
            __extracted_from___class2asset_instance_49(
                remapped_class_instance,
                fallback_property,
                "external_id",
                class_,
            )
        except Exception:
            __extracted_from___class2asset_instance_56(fallback_property, class_, remapped_class_instance)
    # This should not be the use case however to still have name of the object we are using
    # fallback property here as well (typically identifier)
    if "name" in missing_properties:
        try:
            __extracted_from___class2asset_instance_49(remapped_class_instance, fallback_property, "name", class_)
        except Exception:
            __extracted_from___class2asset_instance_56(fallback_property, class_, remapped_class_instance)
    # If object is expected to have parent, but parent is not provided, it is added to orphanage
    # This is typically sign of objects not following proposed ontology/data model/schema
    if "parent_external_id" in missing_properties and orphanage_asset_external_id:
        remapped_class_instance["parent_external_id"] = orphanage_asset_external_id

    if "name" in remapped_class_instance and remapped_class_instance["name"] == "":
        remapped_class_instance["name"] = empty_name_default

    # To maintain shape across of all assets of specific type we are adding missing metadata
    # keys as empty strings, this was request by Statnett
    # Generally this is bad practice, but more of a workaround of their bad data
    if missing_metadata and add_missing_metadata:
        msg = f"Adding missing metadata keys with values set to empty string for {class_}"
        msg += f" instance <{remapped_class_instance['identifier']}>. "
        logging.debug(msg)
        for key in missing_metadata:
            remapped_class_instance[key] = ""
            logging.debug(f"\tKey {key} added to <{remapped_class_instance['identifier']}> metadata!")

    asset_instance = AssetTemplate(
        **remapped_class_instance, external_id_prefix=external_id_prefix, data_set_id=data_set_id
    )

    # Removing field external_id_prefix from asset instance dictionary as it is only
    # convenience field for external_id and parent_external_id update in AssetTemplate
    return asset_instance.dict(exclude={"external_id_prefix"})


# TODO Rename this here and in `__class2asset_instance`
def __extracted_from___class2asset_instance_49(remapped_class_instance, fallback_property, arg2, class_):
    remapped_class_instance[arg2] = remapped_class_instance[fallback_property]
    msg = f"Missing external_id for {class_} instance <{remapped_class_instance['identifier']}>. "
    msg += f"Using value <{remapped_class_instance[fallback_property]}> provided "
    msg += f"by property <{fallback_property}>!"

    logging.debug(msg)


# TODO Rename this here and in `__class2asset_instance`
def __extracted_from___class2asset_instance_56(fallback_property, class_, remapped_class_instance):
    msg = f"Fallback property <{fallback_property}> not found for {class_} "
    msg += f"instance <{remapped_class_instance['identifier']}>."
    logging.error(msg)
    raise ValueError(msg)


def _list2dict(class_instance: list) -> dict:
    """Converting list of class instance properties and values to dictionary

    Parameters
    ----------
    class_instance : list
        Class instance properties and values originating from RDF as list of tuples

    Returns
    -------
    dict
        Class instance properties and values as dictionary
    """

    class_instance_dict = {}
    for property_value_pair in class_instance:
        property_ = remove_namespace(property_value_pair[0])

        # Remove namespace from URIRef values, otherwise convert Literal to string
        # ideally this should react upon property type provided in sheet
        # however Assets only support string values
        value = (
            remove_namespace(property_value_pair[1])
            if isinstance(property_value_pair[1], URIRef)
            else str(property_value_pair[1])
        )

        if property_ in class_instance_dict and value not in class_instance_dict[property_]:
            class_instance_dict[property_] = (
                class_instance_dict[property_] + [value]
                if isinstance(class_instance_dict[property_], list)
                else [class_instance_dict[property_], value]
            )
        else:
            class_instance_dict[property_] = value

    return class_instance_dict


def rdf2assets(
    graph_store: NeatGraphStore,
    transformation_rules: TransformationRules,
    stop_on_exception: bool = False,
    use_orphanage: bool = True,
) -> Dict[str, Asset]:
    """Creates assets from RDF graph

    Parameters
    ----------
    graph : Graph
        Graph containing RDF data
    transformation_rules : TransformationRules
        Instance of TransformationRules class containing transformation rules

    Returns
    -------
    Dict[str, Asset]
        Dictionary of assets with external_id as key
    """

    orphanage_asset_external_id = (
        f"{transformation_rules.metadata.externalIdPrefix or ''}orphanage-{transformation_rules.metadata.data_set_id}"
    )

    graph = graph_store.get_graph()
    # Step 1: Create rdf to asset property mapping
    logging.info("Generating rdf to asset property mapping")
    asset_class_mapping = _define_asset_class_mapping(transformation_rules)

    # Step 4: Get ids of classes
    logging.info("Get ids of instances of classes")
    assets = {}
    class_ids = {
        class_: _get_class_instance_ids(graph, class_, transformation_rules.metadata.namespace)
        for class_ in asset_class_mapping
    }
    # Step 5: Create Assets based on class instances
    logging.info("Create Assets based on class instances")
    for class_ in asset_class_mapping:
        # TODO: Rename class_id to instance_id
        class_ns = transformation_rules.metadata.namespace[class_]
        logging.debug(f"Processing class <{class_ns}> . Number of instances: {len(class_ids[class_])}")
        progress_counter = 0
        # loading all instances into cache
        try:
            query = f"SELECT ?instance ?prop ?value WHERE {{ ?instance rdf:type <{class_ns}> . ?instance ?prop ?value . }} order by ?instance "
            logging.info(query)
            response_df = graph_store.query_to_dataframe(query)
        except Exception as e:
            logging.error(f"Error while loading instances of class <{class_ns}> into cache. Reason: {e}")
            # traceback.print_exc()
            if stop_on_exception:
                raise e
            continue

        grouped_df = response_df.groupby("instance")

        for instance_id, group_df in grouped_df:
            try:
                instance_property_values = group_df.filter(items=["property", "value"]).values.tolist()
                instance_property_values += [(URIRef("http://purl.org/dc/terms/identifier"), URIRef(instance_id))]

                # this will strip namespace from property names and values
                class_instance = _list2dict(instance_property_values)

                # class instance is repaired and converted to asset dictionary
                asset = _class2asset_instance(
                    class_,
                    class_instance,
                    asset_class_mapping[class_],
                    transformation_rules.metadata.data_set_id,
                    orphanage_asset_external_id if use_orphanage else None,  # we need only base external id
                    transformation_rules.metadata.externalIdPrefix or None,
                )

                # adding labels and timestamps
                asset["labels"] = [asset["metadata"]["type"], "non-historic"]
                asset["metadata"]["start_time"] = str(datetime.now(timezone.utc))
                asset["metadata"]["update_time"] = str(datetime.now(timezone.utc))

                # log every 10000 assets
                if progress_counter % 10000 == 0:
                    logging.info(" Next 10000 Assets processed")

                assets[asset["external_id"]] = asset
                progress_counter += 1
            except Exception as ValidationError:
                logging.error(
                    f"Skipping class <{class_}> instance <{remove_namespace(instance_id)}>, reason:\n{ValidationError}\n"
                )
                if stop_on_exception:
                    raise ValidationError

        logging.debug(f"Class <{class_}> processed")

    if orphanage_asset_external_id not in assets:
        _extracted_from_rdf2asset_dictionary_95(orphanage_asset_external_id, transformation_rules, assets)
    logging.info("Assets dictionary created")
    return assets


def rdf2asset_dictionary(
    graph_store: NeatGraphStore,
    transformation_rules: TransformationRules,
    stop_on_exception: bool = False,
    use_orphanage: bool = True,
) -> Dict[str, Asset]:
    warn("'rdf2asset_dictionary' is deprecated, please use 'rdf2assets' instead!")
    logging.warning("'rdf2asset_dictionary' is deprecated, please use 'rdf2assets' instead!")
    return rdf2assets(graph_store, transformation_rules, stop_on_exception, use_orphanage)


# TODO Rename this here and in `rdf2asset_dictionary`
def _extracted_from_rdf2asset_dictionary_95(orphanage_asset_external_id, transformation_rules, assets):
    print(f"Orphanage with external id {orphanage_asset_external_id} not found in asset hierarchy!")
    logging.warning(f"Orphanage with external id {orphanage_asset_external_id} not found in asset hierarchy!")
    logging.warning(f"Adding default orphanage with external id {orphanage_asset_external_id}")

    # updating default external id for orphanage in case there are additional prefix to be added
    ORPHANAGE["external_id"] = orphanage_asset_external_id
    ORPHANAGE["data_set_id"] = transformation_rules.metadata.data_set_id
    ORPHANAGE["metadata"]["start_time"] = str(datetime_utc_now())
    ORPHANAGE["metadata"]["update_time"] = str(datetime_utc_now())
    assets[orphanage_asset_external_id] = ORPHANAGE


def _asset2dict(asset: Asset) -> dict:
    """Return asset as dict representation

    Parameters
    ----------
    asset : Asset
        Instance of Asset class

    Returns
    -------
    dict
        Asset in dict representation
    """

    return {
        "external_id": asset.external_id,
        "name": asset.name,
        "description": asset.description,
        "parent_external_id": asset.parent_external_id,
        "data_set_id": asset.data_set_id,
        "metadata": asset.metadata,
    }


def _flatten_labels(labels: List[Dict[str, str]]) -> set[str]:
    """Flatten labels"""
    return {label["externalId"] for label in labels}


def _is_historic(labels) -> bool:
    """Check if asset is historic"""
    return "historic" in labels


def _categorize_cdf_assets(
    client: CogniteClient, data_set_id: int, partitions: int
) -> Tuple[Optional[pd.DataFrame], dict[str, set]]:
    """Categorize CDF assets

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    data_set_id : int
        Id of data set
    partitions : int
        Number of partitions

    Returns
    -------
    Tuple[Optional[pd.DataFrame], dict]
        CDF assets as pandas dataframe and dictionary with categorized assets
    """
    cdf_assets = client.assets.list(data_set_ids=data_set_id, limit=None, partitions=partitions).to_pandas()

    logging.info(f"Number of assets in CDF {len(cdf_assets)} that have been fetched")

    if cdf_assets.empty:
        return None, {"non-historic": set(), "historic": set()}

    cdf_columns = set(cdf_assets.columns)
    expected_columns = {"external_id", "labels", "parent_external_id", "data_set_id", "name", "description", "metadata"}

    cdf_assets = cdf_assets[list(expected_columns.intersection(cdf_columns))]
    cdf_assets = cdf_assets.where(pd.notnull(cdf_assets), None)
    cdf_assets.labels = cdf_assets.labels.apply(_flatten_labels).values  # type: ignore
    cdf_assets["is_historic"] = cdf_assets.labels.apply(_is_historic).values

    categorized_asset_ids = {
        "historic": set(cdf_assets[cdf_assets.is_historic].external_id.values),
        "non-historic": set(cdf_assets[~cdf_assets.is_historic].external_id.values),
    }

    cdf_assets.drop(["is_historic"], axis=1, inplace=True)
    msg = f"CDF assets categorized into {len(categorized_asset_ids['historic'])} historic"
    msg += f" and {len(categorized_asset_ids['non-historic'])} non-historic assets"
    logging.info(msg)

    return cdf_assets, categorized_asset_ids


def order_assets(assets: dict[str, dict]) -> list[Asset]:
    """Order assets in a way that parent assets are created before child assets

    Parameters
    ----------
    assets : dict[str, dict]
        List of assets to be created

    Returns
    -------
    list[Asset]
        Ordered list of assets
    """
    hierarchy = AssetHierarchy([Asset(**asset) for asset in assets.values()], ignore_orphans=True)
    insert_dct = hierarchy.groupby_parent_xid()
    subtree_count = hierarchy.count_subtree(insert_dct)

    hierarchy = None

    asset_creation_order = pd.DataFrame.from_dict(subtree_count, orient="index", columns=["order"]).sort_values(
        by="order", ascending=False
    )
    asset_creation_order["external_id"] = asset_creation_order.index

    hierarchy = AssetList([Asset(**asset) for asset in assets.values()]).to_pandas()
    hierarchy = hierarchy.where(pd.notnull(hierarchy), None)
    hierarchy = hierarchy.merge(asset_creation_order, left_on="external_id", right_on="external_id")
    hierarchy = hierarchy.sort_values(by="order", ascending=False)
    hierarchy.reset_index(drop=True, inplace=True)
    hierarchy.labels = hierarchy.labels.apply(_flatten_labels)
    hierarchy.drop(["order"], axis=1, inplace=True)

    return [Asset(**row.to_dict()) for _, row in hierarchy.iterrows()]


def _assets_to_create(rdf_assets: dict, asset_ids: set) -> list[Asset]:
    """Return list of assets to be created

    Parameters
    ----------
    rdf_assets : dict
        Dictionary containing assets derived from knowledge graph (RDF)
    asset_ids : set
        Set of asset ids to be created

    Returns
    -------
    list[Asset]
        Ordered list of assets to be created
    """
    start_time = datetime_utc_now()
    if asset_ids:
        logging.info("Wrangling assets to be created into their final form")
        ordered_assets = order_assets({external_id: rdf_assets[external_id] for external_id in asset_ids})

        logging.info(f"Wrangling completed in {(datetime_utc_now() - start_time).seconds} seconds")
        return ordered_assets
    return []


def _assets_to_update(
    rdf_assets: dict,
    cdf_assets: pd.DataFrame,
    asset_ids: set,
    exclude_paths: list = EXCLUDE_PATHS,
    stop_on_exception: bool = False,
) -> tuple[list[Asset], dict[str, dict]]:
    """Return list of assets to be updated

    Parameters
    ----------
    rdf_assets : dict
        Dictionary containing assets derived from knowledge graph (RDF)
    cdf_assets : pd.DataFrame
        Dataframe containing assets from CDF
    asset_ids : set
        Candidate assets to be updated
    exclude_paths : list, optional
        Paths not to be checked when diffing rdf and cdf assets, by default EXCLUDE_PATHS
    stop_on_exception: bool, optional
        Whether to stop on exception or not, by default False

    Returns
    -------
    tuple[list[Asset], dict[str, dict]]
        List of assets to be updated and detailed report of changes per asset
    """

    start_time = datetime_utc_now()
    assets = []
    report = {}
    if not asset_ids:
        return [], {}
    logging.info("Wrangling assets to be updated into their final form")
    cdf_asset_subset = {
        row["external_id"]: row
        for row in cdf_assets[cdf_assets["external_id"].isin(asset_ids)].to_dict(orient="records")
    }
    for external_id in asset_ids:
        cdf_asset = cdf_asset_subset[external_id]
        diffing_result = DeepDiff(cdf_asset, rdf_assets[external_id], exclude_paths=exclude_paths)

        if "parent_external_id" in diffing_result.affected_root_keys:
            msg = f"Asset <{external_id}> is changing its parent from <{cdf_asset['parent_external_id']}>"
            msg += f" to <{rdf_assets[external_id]['parent_external_id']}>! This is not allowed!"
            if stop_on_exception:
                logging.error(msg)
                raise ValueError(msg)
            else:
                msg += " Skipping update of this asset!"
                logging.warning(msg)
                warnings.warn(
                    msg,
                    stacklevel=2,
                )
                continue

        if diffing_result and "root['metadata']['active']" not in diffing_result.affected_paths:
            asset = Asset(**rdf_assets[external_id])
            try:
                asset.metadata["start_time"] = cdf_asset[external_id]["metadata"]["start_time"]
            except KeyError:
                asset.metadata["start_time"] = str(datetime.now(timezone.utc))
            asset.metadata["update_time"] = str(datetime.now(timezone.utc))
            assets.append(asset)

            report[external_id] = dict(diffing_result)

    logging.info(f"Wrangling of {len(assets)} completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return assets, report


def _assets_to_resurrect(rdf_assets: dict, cdf_assets: pd.DataFrame, asset_ids: set) -> list[Asset]:
    """Returns list of assets to be resurrected

    Parameters
    ----------
    rdf_assets : dict
        Dictionary containing assets derived from knowledge graph (RDF)
    cdf_assets : pd.DataFrame
        Dataframe containing assets from CDF
    asset_ids : set
        Set of asset ids to be resurrected

    Returns
    -------
    list[Asset]
        List of assets to be resurrected
    """
    start_time = datetime_utc_now()
    assets = []
    if not asset_ids:
        return []
    logging.info("Wrangling assets to be resurrected into their final form")
    cdf_asset_subset = {
        row["external_id"]: row
        for row in cdf_assets[cdf_assets["external_id"].isin(asset_ids)].to_dict(orient="records")
    }
    for external_id in asset_ids:
        cdf_asset = cdf_asset_subset[external_id]

        asset = Asset(**rdf_assets[external_id])

        try:
            asset.metadata["start_time"] = cdf_asset[external_id]["metadata"]["start_time"]
        except KeyError:
            asset.metadata["start_time"] = str(datetime.now(timezone.utc))
        asset.metadata["update_time"] = str(datetime.now(timezone.utc))
        asset.metadata["resurrection_time"] = str(datetime.now(timezone.utc))
        assets.append(asset)

    logging.info(f"Wrangling of {len(assets)} completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return assets


def _assets_to_decommission(cdf_assets, asset_ids) -> list[Asset]:
    start_time = datetime_utc_now()

    assets = []
    if not asset_ids:
        return []
    logging.info("Wrangling assets to be decommissioned into their final form")
    cdf_asset_subset = {
        row["external_id"]: row
        for row in cdf_assets[cdf_assets["external_id"].isin(asset_ids)].to_dict(orient="records")
    }

    for external_id in asset_ids:
        cdf_asset = cdf_asset_subset[external_id]

        cdf_asset["metadata"]["update_time"] = str(datetime.now(timezone.utc))
        cdf_asset["metadata"].pop("resurrection_time", None)
        cdf_asset["metadata"]["end_time"] = str(datetime.now(timezone.utc))
        cdf_asset["metadata"]["active"] = "false"
        try:
            cdf_asset["labels"].remove("non-historic")
        except KeyError:
            logging.info(f"Asset {external_id} missed label 'non-historic'")
        cdf_asset["labels"].add("historic")

        assets.append(Asset(**cdf_asset))

    logging.info(f"Wrangling completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return assets


def categorize_assets(
    client: CogniteClient,
    rdf_assets: dict,
    data_set_id: int,
    partitions: int = 40,
    stop_on_exception: bool = False,
    return_report: bool = False,
) -> Union[tuple[dict, dict], dict]:
    """Categorize assets on those that are to be created, updated and decommissioned

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    rdf_assets : dict
        Dictionary containing asset external_id - asset pairs
    data_set_id : int
        Dataset id to which assets are to be/are stored
    partitions : int, optional
        Number of partitions to use when fetching assets from CDF, by default 40
    stop_on_exception : bool, optional
        Whether to stop on exception or not, by default False
    return_report : bool, optional
        Whether to report on the diffing results or not, by default False

    Returns
    -------
    Dict[str, list]
        dictionary containing asset category - list of asset pairs
    """
    # TODO: Cache categorized assets somewhere instead of creating them

    cdf_assets, categorized_asset_ids = _categorize_cdf_assets(client, data_set_id, partitions)

    rdf_asset_ids = set(rdf_assets.keys())

    # ids to create
    create_ids = rdf_asset_ids.difference(
        categorized_asset_ids["historic"].union(categorized_asset_ids["non-historic"])
    )

    # ids potentially to update
    update_ids = rdf_asset_ids.intersection(categorized_asset_ids["non-historic"])

    # ids to decommission
    decommission_ids = categorized_asset_ids["non-historic"].difference(rdf_asset_ids)

    # ids to resurrect
    resurrect_ids = categorized_asset_ids["historic"].intersection(rdf_asset_ids)

    logging.info(f"Number of assets to create: { len(create_ids)}")
    logging.info(f"Number of assets to potentially update: { len(update_ids)}")
    logging.info(f"Number of assets to decommission: { len(decommission_ids)}")
    logging.info(f"Number of assets to resurrect: { len(resurrect_ids)}")

    report = {"create": create_ids, "resurrect": resurrect_ids, "decommission": decommission_ids, "update": None}
    categorized_assets = {
        "create": _assets_to_create(rdf_assets, create_ids),
        "update": None,
        "resurrect": _assets_to_resurrect(rdf_assets, cdf_assets, resurrect_ids),
        "decommission": _assets_to_decommission(cdf_assets, decommission_ids),
    }

    categorized_assets["update"], report["update"] = _assets_to_update(
        rdf_assets, cdf_assets, update_ids, stop_on_exception=stop_on_exception
    )

    return (categorized_assets, report) if return_report else categorized_assets


def _micro_batch_push(
    client: CogniteClient, assets: list, batch_size: int = 1000, push_type: str = "update", message: str = "Updated"
):
    """Updates assets in batches of 1000

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    assets : list
        List of assets to be created or updated
    batch_size : int, optional
        Size of batch, by default 1000
    push_type : str, optional
        Type of push, either "update" or "create", by default "update"
    message : str, optional
        Message to logged, by default "Updated"
    """
    total = len(assets)
    counter = 0

    for batch in chunker(assets, batch_size):
        counter += len(batch)
        start_time = datetime_utc_now()
        if push_type == "update":
            client.assets.update(batch)
        elif push_type == "create":
            client.assets.create_hierarchy(batch)
        else:
            logging.info(f"push_type {push_type} not supported")
            raise ValueError(f"push_type {push_type} not supported")

        delta_time = (datetime_utc_now() - start_time).seconds

        msg = f"{message} {counter} of {total} assets, batch processing time: {delta_time:.2f} "
        msg += f"seconds ETC: {delta_time * (total - counter) / (60*batch_size) :.2f} minutes"
        logging.info(msg)


def upload_assets(client: CogniteClient, categorized_assets: Dict[str, list], batch_size: int = 5000):
    """Uploads categorized assets to CDF

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    categorized_assets : Dict[str, list]
        dictionary containing asset category - list of asset pairs
    batch_size : int, optional
        Size of batch, by default 5000
    """
    if batch_size:
        logging.info(f"Uploading assets in batches of {batch_size}")
        if categorized_assets["create"]:
            _micro_batch_push(client, categorized_assets["create"], batch_size, push_type="create", message="Created")

        if categorized_assets["update"]:
            _micro_batch_push(
                client,
                categorized_assets["update"],
                batch_size,
                message="Updated",
            )

        if categorized_assets["resurrect"]:
            _micro_batch_push(
                client,
                categorized_assets["resurrect"],
                batch_size,
                message="Resurrected",
            )

        if categorized_assets["decommission"]:
            _micro_batch_push(
                client,
                categorized_assets["decommission"],
                batch_size,
                message="Decommissioned",
            )

    else:
        logging.info("Batch size not set, pushing all assets to CDF in one go!")
        if categorized_assets["create"]:
            client.assets.create_hierarchy(categorized_assets["create"])

        if categorized_assets["update"]:
            client.assets.create_hierarchy(categorized_assets["update"], upsert=True, upsert_mode="replace")

        if categorized_assets["resurrect"]:
            client.assets.create_hierarchy(categorized_assets["resurrect"], upsert=True, upsert_mode="replace")

        if categorized_assets["decommission"]:
            client.assets.create_hierarchy(categorized_assets["decommission"], upsert=True, upsert_mode="replace")



from __future__ import annotations

import logging
import math
import warnings
from datetime import datetime
from typing import Dict, List, Optional, Self, Union

import pandas as pd
from graphql import GraphQLBoolean, GraphQLFloat, GraphQLInt, GraphQLString
from pydantic import BaseModel, Field, HttpUrl, ValidationError, root_validator, validator
from rdflib import XSD, Literal, Namespace, URIRef

from cognite.neat.core.configuration import PREFIXES, Tables
from cognite.neat.core.data_classes.rules import Entity, RuleType, parse_rule

# mapping of XSD types to Python and GraphQL types
DATA_TYPE_MAPPING = {
    "boolean": {"python": "bool", "GraphQL": GraphQLBoolean},
    "float": {"python": "float", "GraphQL": GraphQLFloat},
    "integer": {"python": "int", "GraphQL": GraphQLInt},
    "nonPositiveInteger": {"python": "int", "GraphQL": GraphQLInt},
    "nonNegativeInteger": {"python": "int", "GraphQL": GraphQLInt},
    "negativeInteger": {"python": "int", "GraphQL": GraphQLInt},
    "long": {"python": "int", "GraphQL": GraphQLInt},
    "string": {"python": "str", "GraphQL": GraphQLString},
    "anyURI": {"python": "str", "GraphQL": GraphQLString},
    "normalizedString": {"python": "str", "GraphQL": GraphQLString},
    "token": {"python": "str", "GraphQL": GraphQLString},
}
METADATA_VALUE_MAX_LENGTH = 5120


class URL(BaseModel):
    url: HttpUrl


class Resource(BaseModel):
    # Solution model
    class_name: str = Field(alias="Class")
    description: str = Field(alias="Description", default="")

    # Solution CDF resource
    cdf_resource_type: str = Field(alias="Resource Type")

    # Solution model advance configuration
    deprecated: bool = Field(default=False)
    deprecation_date: Optional[datetime] = Field(alias="deprecationDate", default=None)
    replaced_by: str = Field(alias="replacedBy", default=None)
    similar_to: HttpUrl = Field(alias="similarTo", default=None)
    similarity_score: float = Field(alias="similarityScore", default=None)
    equal_to: HttpUrl = Field(alias="equalTo", default=None)

    @validator(
        "deprecated",
        "deprecation_date",
        "replaced_by",
        "similar_to",
        "similarity_score",
        "equal_to",
        "description",
        pre=True,
    )
    def replace_float_nan_with_default(cls, value, field):
        if isinstance(value, float) and math.isnan(value):
            return field.default
        return value


class Class(Resource):
    # Solution model
    parent_class: str = Field(alias="Parent Class", default=None)

    # Solution CDF resource
    parent_asset: str = Field(alias="Parent Asset", default=None)

    @validator("parent_class", "parent_asset", pre=True)
    def replace_float_nan_with_default(cls, value, field):
        if isinstance(value, float) and math.isnan(value):
            return field.default
        return value


class Property(Resource):
    # Solution model
    property_name: str = Field(alias="Property")
    expected_value_type: str = Field(alias="Type")
    min_count: Optional[int] = Field(alias="Min Count", default=0)
    max_count: Optional[int] = Field(alias="Max Count", default=None)

    # OWL property
    property_type: str = "DatatypeProperty"

    # Solution CDF resource
    resource_type_property: List[str] = Field(alias="Resource Type Property", default=None)
    source_type: str = Field(alias="Relationship Source Type", default="Asset")
    target_type: str = Field(alias="Relationship Target Type", default="Asset")
    label: str = Field(alias="Relationship Label", default=None)
    relationship_external_id_rule: str = Field(alias="Relationship ExternalID Rule", default=None)

    # Transformation rule (domain to solution)
    rule_type: RuleType = Field(alias="Rule Type", default=None)
    rule: str = Field(alias="Rule", default=None)
    skip_rule: bool = Field(alias="Skip", default=False)

    # Specialization of cdf_resource_type to allow definition of both
    # Asset and Relationship at the same time
    cdf_resource_type: list[str] = Field(alias="Resource Type")

    @property
    def is_raw_lookup(self) -> bool:
        return self.rule_type == RuleType.rawlookup

    @validator(
        "max_count",
        "min_count",
        "target_type",
        "source_type",
        "label",
        "relationship_external_id_rule",
        "resource_type_property",
        "skip_rule",
        "rule",
        "rule_type",
        pre=True,
    )
    def replace_float_nan_with_default(cls, value, field):
        if isinstance(value, float) and math.isnan(value):
            return field.default
        return value

    @validator("rule_type", pre=True)
    def to_lowercase(cls, value):
        return value.casefold() if value else value

    @validator("skip_rule", pre=True)
    def from_string(cls, value):
        if isinstance(value, str):
            return value.casefold() in {"true", "skip", "yes", "y"}
        return value

    @validator("rule")
    def is_valid_rule(cls, value, values):
        if rule_type := values.get("rule_type"):
            _ = parse_rule(value, rule_type)
        return value

    @validator("resource_type_property", pre=True)
    def split_str(cls, v):
        if v:
            return [v.strip() for v in v.split(",")] if "," in v else [v]

    @validator("label")
    def set_relationship_label(cls, value, values):
        if "Relationship" in values.get("cdf_resource_type") and not value:
            return values.get("property_name")
        return value

    @validator("cdf_resource_type", pre=True)
    def to_list_if_comma(cls, value, field):
        if isinstance(value, str):
            if value:
                return value.replace(", ", ",").split(",")
            if field.default is None:
                return None
        return value

    @validator("property_type", pre=True, always=True)
    def set_property_type(cls, value, values):
        if values["expected_value_type"] in DATA_TYPE_MAPPING.keys():
            return "DatatypeProperty"
        else:
            return "ObjectProperty"

    @validator("skip_rule", pre=True, always=True)
    def no_rule(cls, value, values):
        if values.get("rule_type") is None:
            return True
        else:
            return value


class Metadata(BaseModel):
    prefix: str = Field(alias="shortName")
    namespace: Namespace = None
    version: str
    isCurrentVersion: bool = True
    created: datetime
    updated: datetime = Field(default_factory=lambda: datetime.utcnow())
    title: str
    description: Optional[str] = None
    abstract: Optional[str] = None
    creator: str | list[str]
    contributor: Optional[str | list[str]] = None
    rights: Optional[str] = "Restricted for Internal Use of Cognite"
    externalIdPrefix: str = Field(alias="externalIdPrefix", default=None)
    data_set_id: int = Field(alias="dataSetId")

    @validator(
        "externalIdPrefix",
        "contributor",
        "contributor",
        "description",
        "rights",
        pre=True,
    )
    def replace_float_nan_with_default(cls, value, field):
        if isinstance(value, float) and math.isnan(value):
            return field.default
        return value

    @validator("namespace", always=True)
    def set_namespace(cls, value, values):
        if value is None:
            return Namespace(f"http://purl.org/cognite/{values['prefix']}#")

    @validator("creator", "contributor")
    def to_list_if_comma(cls, value, field):
        if isinstance(value, str):
            if value:
                return value.replace(", ", ",").split(",")
            if field.default is None:
                return None
        return value

    class Config:
        allow_population_by_field_name = True

    @classmethod
    def create_from_dataframe(cls, raw_dfs) -> Self:
        expected_tables = Tables.as_set()
        if missing_tables := (expected_tables - set(raw_dfs)):
            raise ValueError(f"Missing the following tables {', '.join(missing_tables)}")
        return Metadata(**dict(zip(raw_dfs[Tables.metadata][0], raw_dfs[Tables.metadata][1])))


class Prefixes(BaseModel):
    prefixes: Dict[str, Namespace] = PREFIXES

    @staticmethod
    def create_from_dataframe(raw_dfs: pd.DataFrame) -> dict[str, Namespace]:
        prefixes = {}
        for i, row in raw_dfs.iterrows():
            try:
                url = URL(url=row["URI"]).url
                prefixes[row["Prefix"]] = Namespace(url)
            except ValueError as e:
                msg = f"Prefix <{row['Prefix']}> has invalid URL: <{row['URI']}> fix this in Prefixes sheet at the row {i + 2} in the rule file!"
                logging.error(msg)
                raise ValueError(msg) from e

        return prefixes


class Instance(BaseModel):
    """Class deals with instances of classes in the data model"""

    instance: URIRef = Field(alias="Instance", default=None)
    property_: URIRef = Field(alias="Property", default=None)
    value: Union[Literal, URIRef] = Field(alias="Value", default=None)
    namespace: URIRef
    prefixes: Dict[str, Namespace]

    @staticmethod
    def get_value(value, prefixes) -> URIRef | Literal:
        try:
            url = URL(url=value).url
            return URIRef(url)
        except ValidationError:
            try:
                entity = Entity.from_string(value)
                return URIRef(prefixes[entity.prefix][entity.name])
            except ValueError:
                return value

    @root_validator(pre=True)
    def convert_values(cls, values: dict):
        # we expect to read Excel sheet which contains naming convention of column
        # 'Instance', 'Property', 'Value', if that's not the case we should raise error
        if not {"Instance", "Property", "Value"}.issubset(set(values.keys())):
            raise TypeError("We only support inputs from the transformation rule Excel sheet!!!")

        namespace = values["namespace"]
        prefixes = values["prefixes"]

        values["Instance"] = cls.get_value(values["Instance"], prefixes)
        values["Instance"] = (
            values["Instance"] if isinstance(values["Instance"], URIRef) else URIRef(namespace[values["Instance"]])
        )

        values["Property"] = cls.get_value(values["Property"], prefixes)
        values["Property"] = (
            values["Property"] if isinstance(values["Property"], URIRef) else URIRef(namespace[values["Property"]])
        )

        values["Value"] = cls.get_value(values["Value"], prefixes)

        if not isinstance(values["Value"], URIRef):
            datatype = (
                XSD.integer if cls.isint(values["Value"]) else XSD.float if cls.isfloat(values["Value"]) else XSD.string
            )
            values["Value"] = Literal(values["Value"], datatype=datatype)

        return values

    @staticmethod
    def isfloat(x):
        try:
            _ = float(x)
        except (TypeError, ValueError):
            return False
        else:
            return True

    @staticmethod
    def isint(x):
        try:
            a = float(x)
            b = int(a)
        except (TypeError, ValueError):
            return False
        else:
            return a == b


class TransformationRules(BaseModel):
    metadata: Metadata
    classes: dict[str, Class]
    properties: dict[str, Property]
    prefixes: dict[str, Namespace]
    instances: list[tuple] = None

    @property
    def raw_tables(self) -> list[str]:
        return list(
            {
                parse_rule(rule.rule, RuleType.rawlookup).table.name
                for rule in self.properties.values()
                if rule.is_raw_lookup
            }
        )

    @validator("properties", each_item=True)
    def is_valid_mapping(cls, value, values):
        if classes := values.get("classes"):
            if value.class_name not in classes:
                raise ValueError(f"invalid mapping rule, {value.class_name} does not exist ")
        return value

    @validator("properties", each_item=True)
    def value_type_exist(cls, value, values):
        if classes := values.get("classes"):
            if value.property_type == "ObjectProperty" and value.expected_value_type not in classes:
                msg = f"Property <{value.property_name}> defined for class <{value.class_name}> has "
                msg += f"value type <{value.expected_value_type}> which is not defined!"
                raise ValueError(msg)
        return value

    @validator("properties", each_item=True)
    def add_missing_label(cls, value):
        "Add label if missing for relationships"
        if value.label is None and "Relationship" in value.cdf_resource_type:
            value.label = value.property_name
        return value

    @validator("properties")
    def is_type_defined_as_object(cls, value):
        "Checks if property expected value type is defined as object"
        defined_objects = {property_.class_name for property_ in value.values()}

        if undefined_objects := [
            id
            for id, property_ in value.items()
            if property_.property_type == "ObjectProperty" and property_.expected_value_type not in defined_objects
        ]:
            msg = [
                f"\nProperty at {id} has type <{value[id].expected_value_type}> for which mapping is not defined!"
                for id in undefined_objects
            ]
            raise ValueError("".join(msg))
        return value

    @validator("properties", each_item=True)
    def class_property_exist(cls, value, values):
        if classes := values.get("classes"):
            if value.class_name not in classes:
                raise ValueError(
                    f"Property <{value.property_name}> defined for non-existing class <{value.class_name}>!"
                )
        return value

    @validator("prefixes")
    def add_data_model_prefix_namespace(cls, value, values):
        value[values["metadata"].prefix] = values["metadata"].namespace
        return value

    def get_labels(self) -> set[str]:
        """Return CDF labels for classes and relationships."""
        class_labels = {class_.class_name for class_ in self.classes.values()}

        property_labels = {property_.property_name for property_ in self.properties.values()}

        relationship_labels = {
            rule.label for rule in self.properties.values() if "Relationship" in rule.cdf_resource_type
        }

        return class_labels.union(relationship_labels).union(property_labels)

    def get_defined_classes(self) -> set[str]:
        """Returns classes that have been defined in the data model."""
        return {property.class_name for property in self.properties.values()}

    def get_classes_with_properties(self) -> dict[str, Property]:
        """Returns classes that have been defined in the data model."""
        # TODO: Do not particularly like method name, find something more suitable
        class_property_pairs = {}

        for property_ in self.properties.values():
            class_ = property_.class_name
            if class_ in class_property_pairs:
                class_property_pairs[class_] += [property_]
            else:
                class_property_pairs[class_] = [property_]

        return class_property_pairs

    def check_data_model_definitions(self):
        """Check if data model definitions are valid."""
        issues = set()
        for class_, properties in self.get_classes_with_properties().items():
            analyzed_properties = []
            for property in properties:
                if property.property_name not in analyzed_properties:
                    analyzed_properties.append(property.property_name)
                else:
                    issues.add(f"Property {property.property_name} of class {class_} has been defined more than once!")
        return issues

    def reduce_data_model(self, desired_classes: set, skip_validation: bool = False) -> TransformationRules:
        """Reduce the data model to only include desired classes and their properties.

        Parameters
        ----------
        desired_classes : set
            Desired classes to include in the reduced data model
        skip_validation : bool
            To skip underlying pydantic validation, by default False

        Returns
        -------
        TransformationRules
            Instance of TransformationRules

        Notes
        -----
        It is fine to skip validation since we are deriving the reduced data model from data
        model (i.e. TransformationRules) which has already been validated.
        """

        defined_classes = self.get_defined_classes()
        possible_classes = defined_classes.intersection(desired_classes)
        impossible_classes = desired_classes - possible_classes

        if not possible_classes:
            logging.error("None of the desired classes are defined in the data model!")
            raise ValueError("None of the desired classes are defined in the data model!")

        if impossible_classes:
            logging.warning(f"Could not find the following classes defined in the data model: {impossible_classes}")
            warnings.warn(
                f"Could not find the following classes defined in the data model: {impossible_classes}", stacklevel=2
            )

        reduced_data_model = {
            "metadata": self.metadata,
            "prefixes": self.prefixes,
            "classes": {},
            "properties": {},
            "instances": self.instances,
        }

        logging.info(f"Reducing data model to only include the following classes: {possible_classes}")
        for class_ in possible_classes:
            reduced_data_model["classes"][class_] = self.classes[class_]

        for id_, property_definition in self.properties.items():
            if property_definition.class_name in possible_classes:
                reduced_data_model["properties"][id_] = property_definition

        if skip_validation:
            return TransformationRules.construct(**reduced_data_model)
        else:
            return TransformationRules(**reduced_data_model)

    def to_dataframe(self) -> Dict[str, pd.DataFrame]:
        """Represent data model as a dictionary of data frames, where each data frame
        represents properties defined for a given class.

        Returns
        -------
        Dict[str, pd.DataFrame]
            Simplified representation of the data model
        """

        data_model = {}

        defined_classes = self.get_classes_with_properties()

        for class_ in defined_classes:
            properties = {}
            for property_ in defined_classes[class_]:
                if property_.property_name not in properties:
                    properties[property_.property_name] = {
                        "property_type": property_.property_type,
                        "value_type": property_.expected_value_type,
                        "min_count": property_.min_count,
                        "max_count": property_.max_count,
                    }

            data_model[class_] = pd.DataFrame(properties).T

        return data_model

    def get_class_linkage(self) -> pd.DataFrame:
        """Returns a dataframe with the class linkage of the data model."""

        class_linkage = pd.DataFrame(columns=["source_class", "target_class", "connecting_property", "max_occurrence"])
        for property_ in self.properties.values():
            if property_.property_type == "ObjectProperty":
                new_row = pd.Series(
                    {
                        "source_class": property_.class_name,
                        "target_class": property_.expected_value_type,
                        "connecting_property": property_.property_name,
                        "max_occurrence": property_.max_count,
                    }
                )
                class_linkage = pd.concat([class_linkage, new_row.to_frame().T], ignore_index=True)

        class_linkage.drop_duplicates(inplace=True)

        return class_linkage

    def get_connected_classes(self) -> set:
        """Return a set of classes that are connected to other classes."""
        class_linkage = self.get_class_linkage()
        return set(class_linkage.source_class.values).union(set(class_linkage.target_class.values))

    def get_disconnected_classes(self):
        """Return a set of classes that are disconnected (i.e. isolated) from other classes."""
        return self.get_defined_classes() - self.get_connected_classes()

    def get_symmetric_pairs(self) -> list:
        """Returns a list of pairs of symmetrically linked classes."""
        # TODO: Find better name for this method

        class_linkage = self.get_class_linkage()
        if class_linkage.empty:
            return []

        sym_pairs = set()
        for _, row in class_linkage.iterrows():
            source = row.source_class
            target = row.target_class
            target_targets = class_linkage[class_linkage.source_class == target].target_class.values
            if source in target_targets and (source, target) not in sym_pairs:
                sym_pairs.add((source, target))
        return sym_pairs

    def define_relationships(self, stop_on_exception: bool = False) -> RelationshipDefinitions:
        relationships = {}

        # Unique ids used to check for redefinitions of relationships
        ids = set()

        for row, rule in self.properties.items():
            if "Relationship" in rule.cdf_resource_type:
                relationship = RelationshipDefinition(
                    source_class=rule.class_name,
                    target_class=rule.expected_value_type,
                    property_=rule.property_name,
                    labels=list(
                        set([rule.label, rule.class_name, rule.expected_value_type, "non-historic", rule.property_name])
                    ),
                    target_type=rule.target_type,
                    source_type=rule.source_type,
                    relationship_external_id_rule=rule.relationship_external_id_rule,
                )

                id_ = f"{rule.class_name}({rule.property_name})"
                if id_ in ids:
                    msg = f"Relationship {rule.property_name} redefined at {row} in transformation rules!"
                    if stop_on_exception:
                        logging.error(msg)
                        raise ValueError(msg)
                    else:
                        msg += " Skipping redefinition!"
                        warnings.warn(msg, stacklevel=2)
                        logging.warning(msg)
                else:
                    relationships[row] = relationship
                    ids.add(id_)

        if relationships:
            return RelationshipDefinitions(
                data_set_id=self.metadata.data_set_id,
                prefix=self.metadata.prefix,
                namespace=self.metadata.namespace,
                relationships=relationships,
            )

        msg = "No relationship defined in transformation rule sheet!"
        if stop_on_exception:
            logging.error(msg)
            raise ValueError(msg)
        else:
            warnings.warn(msg, stacklevel=2)
            logging.warning(msg)
            return RelationshipDefinitions(
                data_set_id=self.metadata.data_set_id,
                prefix=self.metadata.prefix,
                namespace=self.metadata.namespace,
                relationships={},
            )

    def get_entity_names(self):
        class_names = set()
        property_names = set()
        for class_, properties in self.to_dataframe().items():
            class_names.add(class_)
            property_names = property_names.union(set(properties.index))
        return class_names.union(property_names)


class AssetClassMapping(BaseModel):
    external_id: str
    name: str
    parent_external_id: Optional[str]
    description: Optional[str]
    metadata: Optional[dict] = {}

    @root_validator(pre=True)
    def create_metadata(cls, values: dict):
        fields = values.keys()

        # adding metadata key in case if it is missing
        values["metadata"] = {} if "metadata" not in values else values["metadata"]

        for field in fields:
            if field not in ["external_id", "name", "parent_external_id", "data_set_id", "metadata", "description"]:
                values["metadata"][field] = ""
        return values


class AssetTemplate(BaseModel):
    """This class is used to validate, repair and wrangle rdf asset dictionary according to the
    expected format of cognite sdk Asset dataclass."""

    external_id_prefix: Optional[str] = None  # convenience field to add prefix to external_ids
    external_id: str
    name: Optional[str]
    parent_external_id: Optional[str]
    metadata: Optional[dict] = {}
    description: Optional[str]
    data_set_id: Optional[int] = None

    @root_validator(pre=True)
    def preprocess_fields(cls, values: dict):
        fields = values.keys()

        # Adding metadata key in case if it is missing
        values["metadata"] = {} if "metadata" not in values else values["metadata"]

        for field in fields:
            # Enrich: adding any field that is not in the list of expected fields to metadata
            if field not in [
                "external_id",
                "name",
                "parent_external_id",
                "data_set_id",
                "metadata",
                "description",
                "external_id_prefix",
            ]:
                values["metadata"][field] = values[field]

            # Repair: in case if name/description is list instead of single value list elements are joined
            elif field in ["name", "description"] and isinstance(values[field], list):
                msg = f"{values['type']} instance {values['identifier']} property {field} "
                msg += f"has multiple values {values[field]}, "
                msg += f"these values will be joined in a single string: {', '.join(values[field])}"
                logging.info(msg)
                values[field] = ", ".join(values[field])[: METADATA_VALUE_MAX_LENGTH - 1]

            # Repair: in case if external_id or parent_external_id are lists, we take the first value
            elif field in ["external_id", "parent_external_id"] and isinstance(values[field], list):
                msg = f"{values['type']} instance {values['identifier']} property {field} "
                msg += f"has multiple values {values[field]}, "
                msg += f"only the first one will be used: {values[field][0]}"
                logging.info(msg)
                values[field] = values[field][0]

        # Setting asset to be by default active
        values["metadata"]["active"] = "true"

        # Handling case when the external_id is not provided by defaulting to the original identifier
        # The original identifier probably has its namespace removed
        if "external_id" not in fields and "identifier" in fields:
            values["external_id"] = values["identifier"]

        return values

    @validator("metadata")
    def to_list_if_comma(cls, value):
        for key, v in value.items():
            if isinstance(v, list):
                value[key] = ", ".join(v)[: METADATA_VALUE_MAX_LENGTH - 1]
        return value

    @validator("metadata")
    def to_str(cls, value):
        for key, v in value.items():
            value[key] = str(v)
        return value

    @validator("external_id", always=True)
    def add_prefix_to_external_id(cls, value, values):
        if values["external_id_prefix"]:
            return values["external_id_prefix"] + value
        else:
            return value

    @validator("parent_external_id")
    def add_prefix_to_parent_external_id(cls, value, values):
        if values["external_id_prefix"]:
            return values["external_id_prefix"] + value
        else:
            return value


class RelationshipDefinition(BaseModel):
    source_class: str
    target_class: str
    property_: str
    labels: Optional[List[str]] = None
    target_type: str = "Asset"
    source_type: str = "Asset"
    relationship_external_id_rule: str = None


class RelationshipDefinitions(BaseModel):
    data_set_id: int
    prefix: str
    namespace: Namespace
    relationships: dict[str, RelationshipDefinition]



"""Methods to transform Excel Sheet to Data Model
"""

from rdflib import DCTERMS, OWL, RDFS, SKOS, XSD
from rdflib.graph import RDF, BNode, Collection, Graph, Literal, Namespace, URIRef

from cognite.neat.core.configuration import PREFIXES
from cognite.neat.core.data_classes.data_model_definitions import DataModelingDefinition
from cognite.neat.core.data_classes.transformation_rules import Class

SHACL = Namespace("http://www.w3.org/ns/shacl#")


def _wrangle_owl_properties(data_model_definition: DataModelingDefinition) -> dict[str, dict[str, set]]:
    """Wrangles the properties of the data model definition into a dictionary for easier conversion to RDF

    Parameters
    ----------
    data_model_definition : DataModelingDefinition
        Instance of DataModelingDefinition to be wrangled

    Returns
    -------
    dict[str, dict[str, set]]
        Dictionary of properties for easier conversion to RDF
    """
    # TODO: Add support for deprecated, replaced_by, similar_to, equal_to
    properties = {}

    for property_ in data_model_definition.properties.values():
        if property_.property_name != "*":
            if property_.property_name not in properties:
                properties[property_.property_name] = {
                    "domain": {data_model_definition.namespace[property_.class_name]},
                    "range": {
                        data_model_definition.namespace[property_.expected_value_type]
                        if property_.property_type == "ObjectProperty"
                        else XSD[property_.expected_value_type]
                    },
                    "property_type": {
                        OWL[property_.property_type],
                    },
                    "description": {property_.description or None},
                    # "equal_to": {property_.equal_to or None},
                    # "similar_to": {property_.similar_to or None},
                    # "replaced_by": {property_.replaced_by or None},
                    # "deprecated": {property_.deprecated or None},
                }
            else:
                properties[property_.property_name]["domain"].add(data_model_definition.namespace[property_.class_name])
                properties[property_.property_name]["range"].add(
                    data_model_definition.namespace[property_.expected_value_type]
                    if property_.property_type == "ObjectProperty"
                    else XSD[property_.expected_value_type]
                )
                properties[property_.property_name]["property_type"].add(OWL[property_.property_type])
                properties[property_.property_name]["description"].add(property_.description or None)
                # properties[property_.property_name]["equal_to"].add(property_.equal_to or None)
                # properties[property_.property_name]["similar_to"].add(property_.similar_to or None)
                # properties[property_.property_name]["replaced_by"].add(property_.replaced_by or None)
                # properties[property_.property_name]["deprecated"].add(property_.deprecated or None)

    return properties


def _add_ontology_metadata(ontology_graph: Graph, data_model_definition: DataModelingDefinition) -> None:
    """Adds metadata to the ontology graph

    Parameters
    ----------
    ontology_graph : Graph
        Ontology graph to add metadata to
    data_model_definition : DataModelingDefinition
        Instance of DataModelingDefinition to add metadata from
    """
    ontology_graph.add(
        (URIRef(data_model_definition.namespace), DCTERMS.hasVersion, Literal(data_model_definition.metadata.version))
    )
    ontology_graph.add(
        (URIRef(data_model_definition.namespace), OWL.versionInfo, Literal(data_model_definition.metadata.version))
    )
    ontology_graph.add(
        (URIRef(data_model_definition.namespace), RDFS.label, Literal(data_model_definition.metadata.title))
    )
    ontology_graph.add(
        (
            URIRef(data_model_definition.namespace),
            DCTERMS.created,
            Literal(data_model_definition.metadata.created, datatype=XSD.dateTime),
        )
    )
    ontology_graph.add(
        (
            URIRef(data_model_definition.namespace),
            DCTERMS.modified,
            Literal(data_model_definition.metadata.updated, datatype=XSD.dateTime),
        )
    )

    if data_model_definition.metadata.rights:
        ontology_graph.add(
            (URIRef(data_model_definition.namespace), DCTERMS.rights, Literal(data_model_definition.metadata.rights))
        )

    if data_model_definition.metadata.description:
        ontology_graph.add(
            (URIRef(data_model_definition.namespace), RDFS.comment, Literal(data_model_definition.metadata.description))
        )

    if data_model_definition.metadata.creator:
        [
            ontology_graph.add((URIRef(data_model_definition.namespace), DCTERMS.creator, Literal(creator)))
            for creator in data_model_definition.metadata.creator
        ]

    if data_model_definition.metadata.contributor:
        [
            ontology_graph.add((URIRef(data_model_definition.namespace), DCTERMS.contributor, Literal(contributor)))
            for contributor in data_model_definition.metadata.contributor
        ]


def _add_owl_property(
    ontology_graph: Graph, data_model_definition: DataModelingDefinition, properties: dict, property_: str
) -> None:
    """Add OWL property to the ontology graph

    Parameters
    ----------
    ontology_graph : Graph
        Ontology graph to add property to
    data_model_definition : DataModelingDefinition
        Instance of DataModelingDefinition to add property from
    properties : dict
        Wrangled properties
    property_ : str
        Specific property to add to the ontology graph which is a key in the properties dict
    """
    ontology_graph.add((data_model_definition.namespace[property_], RDFS.label, Literal(property_)))

    if len(properties[property_]["description"]) == 1:
        ontology_graph.add(
            (
                data_model_definition.namespace[property_],
                RDFS.comment,
                Literal(list(properties[property_]["description"])[0]),
            )
        )
    else:
        print(f"WARNING: <{property_}> has multiple definitions which will be concatenated")
        ontology_graph.add(
            (
                data_model_definition.namespace[property_],
                RDFS.comment,
                Literal("\n".join(list(properties[property_]["description"]))),
            )
        )

    # add description
    if len(properties[property_]["property_type"]) == 1:
        ontology_graph.add(
            (data_model_definition.namespace[property_], RDF.type, list(properties[property_]["property_type"])[0])
        )
    else:
        print(
            f"BAD PRACTICE: Property <{property_}> is of multiple property types: {', '.join(list(properties[property_]['property_type']))}"
        )
        [
            ontology_graph.add((data_model_definition.namespace[property_], RDF.type, property_type))
            for property_type in properties[property_]["property_type"]
        ]

    if len(properties[property_]["domain"]) == 1:
        ontology_graph.add(
            (data_model_definition.namespace[property_], RDFS.domain, list(properties[property_]["domain"])[0])
        )
    else:
        print(
            f"WARNING: Property <{property_}> domain is union of multiple classes: {', '.join(list(properties[property_]['domain']))}"
        )
        b_union = BNode()
        b_domain = BNode()
        ontology_graph.add((data_model_definition.namespace[property_], RDFS.domain, b_domain))
        ontology_graph.add((b_domain, OWL.unionOf, b_union))
        ontology_graph.add((b_domain, RDF.type, OWL.Class))
        _ = Collection(ontology_graph, b_union, list(properties[property_]["domain"]))

    if len(properties[property_]["range"]) == 1:
        ontology_graph.add(
            (data_model_definition.namespace[property_], RDFS.range, list(properties[property_]["range"])[0])
        )
    else:
        print(
            f"WARNING: Property <{property_}> range is union of multiple types: {', '.join(list(properties[property_]['range']))}"
        )
        b_union = BNode()
        b_range = BNode()
        ontology_graph.add((data_model_definition.namespace[property_], RDFS.range, b_range))
        ontology_graph.add((b_range, OWL.unionOf, b_union))
        ontology_graph.add((b_range, RDF.type, OWL.Class))
        _ = Collection(ontology_graph, b_union, list(properties[property_]["range"]))


def _add_owl_class(ontology_graph: Graph, data_model_definition: DataModelingDefinition, class_: Class) -> None:
    """Add OWL class to ontology graph

    Parameters
    ----------
    ontology_graph : Graph
        Ontology graph
    data_model_definition : DataModelingDefinition
        Instance of DataModelingDefinition class
    class_ : Class
        Class to be added to ontology graph
    """
    # TODO: Simplify by only providing namespace instead of data_model_definition
    ontology_graph.add((data_model_definition.namespace[class_.class_name], RDF.type, OWL.Class))
    ontology_graph.add((data_model_definition.namespace[class_.class_name], RDFS.subClassOf, OWL.Thing))

    # add datatype recognition
    ontology_graph.add((data_model_definition.namespace[class_.class_name], RDFS.label, Literal(class_.class_name)))

    if class_.description:
        ontology_graph.add(
            (data_model_definition.namespace[class_.class_name], RDFS.comment, Literal(class_.description))
        )

    if class_.parent_class:
        ontology_graph.add(
            (
                data_model_definition.namespace[class_.class_name],
                RDFS.subClassOf,
                data_model_definition.namespace[class_.parent_class],
            )
        )

    if class_.deprecated:
        ontology_graph.add(
            (
                data_model_definition.namespace[class_.class_name],
                OWL.deprecated,
                Literal(class_.deprecated, datatype=XSD.boolean),
            )
        )

    if class_.replaced_by:
        ontology_graph.add(
            (
                data_model_definition.namespace[class_.class_name],
                DCTERMS.isReplacedBy,
                data_model_definition.namespace[class_.replaced_by],
            )
        )

    if class_.similar_to:
        ontology_graph.add((data_model_definition.namespace[class_.class_name], SKOS.exactMatch, class_.similar_to))

    if class_.equal_to:
        ontology_graph.add((data_model_definition.namespace[class_.class_name], OWL.equivalentClass, class_.equal_to))


def data_model_definition2owl(
    data_model_definition: DataModelingDefinition,
    prefixes: dict[str, Namespace] = PREFIXES,
    owl_graph: Graph = None,
) -> Graph:
    """Generate OWL ontology from data model definitions

    Parameters
    ----------
    data_model_definition : DataModelingDefinition
        Instance of DataModelingDefinition class
    prefixes : dict[str, Namespace], optional
        Prefixes followed with associated namespaces, by default PREFIXES
    owl_graph : Graph, optional
        ontology graph in case if it is being extended, by default None

    Returns
    -------
    Graph
        OWL ontology based on data model definitions
    """

    if not owl_graph:
        owl_graph = Graph()
        # Bind Data Model namespace and prefix
        owl_graph.bind(data_model_definition.prefix, data_model_definition.namespace)
        # Bind other prefixes and namespaces
        for prefix, namespace in prefixes.items():
            owl_graph.bind(prefix, namespace)

    # add metadata to the graph
    _add_ontology_metadata(owl_graph, data_model_definition)

    # add OWL ontology classes to the graph
    for class_ in data_model_definition.classes.values():
        _add_owl_class(owl_graph, data_model_definition, class_)

    # add OWL ontology properties to the graph
    if owl_properties := _wrangle_owl_properties(data_model_definition):
        for property_ in owl_properties:
            _add_owl_property(owl_graph, data_model_definition, owl_properties, property_)

    return owl_graph


def data_model_definition2shacl(
    data_model_definition: DataModelingDefinition,
    prefixes: dict[str, Namespace] = PREFIXES,
    shacl_graph: Graph = None,
) -> Graph:
    if not shacl_graph:
        shacl_graph = Graph()
        # Bind Data Model namespace and prefix
        shacl_graph.bind(data_model_definition.prefix, data_model_definition.namespace)
        # Bind other prefixes and namespaces
        for prefix, namespace in prefixes.items():
            shacl_graph.bind(prefix, namespace)

    node_shapes = {}
    for shacl_constraint in data_model_definition.properties.values():
        property_name = shacl_constraint.property_name
        if property_name != "*":
            class_name = shacl_constraint.class_name
            node_shape_name = f"{class_name}Shape"

            # adds node shape to shacl graph
            if node_shape_name not in node_shapes:
                node_shapes[node_shape_name] = {}
                shacl_graph.add((data_model_definition.namespace[node_shape_name], RDF.type, SHACL.NodeShape))
                shacl_graph.add(
                    (
                        data_model_definition.namespace[node_shape_name],
                        RDFS.comment,
                        Literal(data_model_definition.classes[class_name].description),
                    )
                )
                shacl_graph.add(
                    (
                        data_model_definition.namespace[node_shape_name],
                        SHACL.targetClass,
                        data_model_definition.namespace[class_name],
                    )
                )

            # adds property shape to shacl graph
            property_shape_node = BNode()
            shacl_graph.add((data_model_definition.namespace[node_shape_name], SHACL.property, property_shape_node))
            shacl_graph.add((property_shape_node, SHACL.path, data_model_definition.namespace[property_name]))
            shacl_graph.add((property_shape_node, SHACL.name, Literal(property_name)))
            if shacl_constraint.property_type == "ObjectProperty":
                shacl_graph.add((property_shape_node, SHACL.nodeKind, SHACL.IRI))
                shacl_graph.add(
                    (
                        property_shape_node,
                        SHACL.node,
                        data_model_definition.namespace[f"{shacl_constraint.expected_value_type}Shape"],
                    )
                )
            else:
                shacl_graph.add((property_shape_node, SHACL.nodeKind, SHACL.Literal))
                shacl_graph.add((property_shape_node, SHACL.datatype, XSD[shacl_constraint.expected_value_type]))

            if shacl_constraint.min_count:
                shacl_graph.add((property_shape_node, SHACL.minCount, Literal(shacl_constraint.min_count)))
            if shacl_constraint.max_count:
                shacl_graph.add((property_shape_node, SHACL.maxCount, Literal(shacl_constraint.max_count)))

    return shacl_graph



from .config import Config
from .transformation_rules import (
    AssetClassMapping,
    AssetTemplate,
    Class,
    Instance,
    Metadata,
    Prefixes,
    Property,
    Resource,
    TransformationRules,
)

__all__ = [
    "Config",
    "TransformationRules",
    "Class",
    "Property",
    "Resource",
    "Metadata",
    "Prefixes",
    "AssetClassMapping",
    "AssetTemplate",
    "Instance",
]



import logging

import pandas as pd

from cognite.neat.core.configuration import Tables
from cognite.neat.core.data_classes import Instance, Metadata, Prefixes, TransformationRules


def parse_transformation_rules(raw_dfs: dict[str, pd.DataFrame]) -> TransformationRules:
    metadata = Metadata.create_from_dataframe(raw_dfs)
    prefixes = Prefixes.create_from_dataframe(raw_dfs[Tables.prefixes])
    instances = parse_instances(raw_dfs, metadata, prefixes) if Tables.instances in raw_dfs else None

    return TransformationRules(
        prefixes=prefixes,
        metadata=metadata,
        classes={class_.get("Class"): class_ for class_ in raw_dfs[Tables.classes].to_dict(orient="records")},
        properties={
            f"row {i+3}": property_ for i, property_ in enumerate(raw_dfs[Tables.properties].to_dict(orient="records"))
        },
        instances=instances,
    )


def parse_instances(raw_dfs: dict[str, pd.DataFrame], metadata: Metadata, prefixes: Prefixes) -> list[tuple]:
    expected_tables = Tables.as_set()
    if missing_tables := (expected_tables - set(raw_dfs)):
        raise ValueError(f"Missing the following tables {', '.join(missing_tables)}")

    prefixes[metadata.prefix] = metadata.namespace

    instances = []
    for row_no, row in raw_dfs[Tables.instances].iterrows():
        try:
            triple = Instance(**row.to_dict(), namespace=metadata.namespace, prefixes=prefixes)
            instances += [(triple.instance, triple.property_, triple.value)]
        except Exception:
            msg = f"Skipping row <{row_no + 3}> in Instance sheet\nReason: prefix in Property or Value column not defined!\n"
            print(msg)
            logging.info(msg)

    return instances



import logging
import time
from pathlib import Path

import requests
from cognite.client import CogniteClient

from cognite.neat.core.data_classes.transformation_rules import TransformationRules
from cognite.neat.core.loader.graph_store import NeatGraphStore
from cognite.neat.core.workflow.base import BaseWorkflow, FlowMessage


class GraphDbImportNeatWorkflow(BaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client, [])
        self.dataset_id: int = 0
        self.current_step: str = None
        self.source_graph: NeatGraphStore = None
        self.transformation_rules: TransformationRules = None
        self.stop_on_error = False
        self.extra_triplets = []

    def step_download_files_from_cdf(self, flow_msg: FlowMessage = None):
        # Download files from CDF
        logging.info("Downloading files from CDF")
        rdf_file_external_id = self.get_config_item("rdf_file_import.cdf_exteranl_id").value
        local_import_folder = self.get_config_item("rdf_file_import.shared_folder").value
        self.cdf_client.files.download(local_import_folder, external_id=rdf_file_external_id)

    def step_create_repository(self, flow_msg: FlowMessage = None):
        self.source_graph_api_url = self.get_config_item("source_rdf_store.api_root_url").value
        # Load rules from file or remote location
        logging.info("CREATEING REPO 2")
        repo_name = self.get_config_item("graphdb_repo_name").value
        req = {
            "id": repo_name,
            "params": {
                "queryTimeout": {"name": "queryTimeout", "label": "Query timeout (seconds)", "value": 30},
                "cacheSelectNodes": {"name": "cacheSelectNodes", "label": "Cache select nodes", "value": "true"},
                "rdfsSubClassReasoning": {
                    "name": "rdfsSubClassReasoning",
                    "label": "RDFS subClass reasoning",
                    "value": "true",
                },
                "validationEnabled": {
                    "name": "validationEnabled",
                    "label": "Enable the SHACL validation",
                    "value": "true",
                },
                "ftsStringLiteralsIndex": {
                    "name": "ftsStringLiteralsIndex",
                    "label": "FTS index for xsd:string literals",
                    "value": "default",
                },
                "shapesGraph": {
                    "name": "shapesGraph",
                    "label": "Named graphs for SHACL shapes",
                    "value": "http://rdf4j.org/schema/rdf4j#SHACLShapeGraph",
                },
                "parallelValidation": {
                    "name": "parallelValidation",
                    "label": "Run parallel validation",
                    "value": "true",
                },
                "title": {"name": "title", "label": "Repository description", "value": ""},
                "checkForInconsistencies": {
                    "name": "checkForInconsistencies",
                    "label": "Enable consistency checks",
                    "value": "false",
                },
                "performanceLogging": {
                    "name": "performanceLogging",
                    "label": "Log the execution time per shape",
                    "value": "false",
                },
                "disableSameAs": {"name": "disableSameAs", "label": "Disable owl:sameAs", "value": "true"},
                "ftsIrisIndex": {
                    "name": "ftsIrisIndex",
                    "label": "FTS index for full-text indexing of IRIs",
                    "value": "none",
                },
                "entityIndexSize": {"name": "entityIndexSize", "label": "Entity index size", "value": "10000000"},
                "dashDataShapes": {"name": "dashDataShapes", "label": "DASH data shapes extensions", "value": "true"},
                "queryLimitResults": {"name": "queryLimitResults", "label": "Limit query results", "value": 0},
                "throwQueryEvaluationExceptionOnTimeout": {
                    "name": "throwQueryEvaluationExceptionOnTimeout",
                    "label": "Throw exception on query timeout",
                    "value": "false",
                },
                "id": {"name": "id", "label": "Repository ID", "value": "repo-test"},
                "storageFolder": {"name": "storageFolder", "label": "Storage folder", "value": "storage"},
                "validationResultsLimitPerConstraint": {
                    "name": "validationResultsLimitPerConstraint",
                    "label": "Validation results limit per constraint",
                    "value": 1000,
                },
                "enablePredicateList": {
                    "name": "enablePredicateList",
                    "label": "Enable predicate list index",
                    "value": "true",
                },
                "transactionalValidationLimit": {
                    "name": "transactionalValidationLimit",
                    "label": "Transactional validation limit",
                    "value": "500000",
                },
                "ftsIndexes": {
                    "name": "ftsIndexes",
                    "label": "FTS indexes to build (comma delimited)",
                    "value": "default, iri",
                },
                "logValidationPlans": {
                    "name": "logValidationPlans",
                    "label": "Log the executed validation plans",
                    "value": "false",
                },
                "imports": {"name": "imports", "label": "Imported RDF files(';' delimited)", "value": ""},
                "inMemoryLiteralProperties": {
                    "name": "inMemoryLiteralProperties",
                    "label": "Cache literal language tags",
                    "value": "true",
                },
                "isShacl": {"name": "isShacl", "label": "Enable SHACL validation", "value": "false"},
                "ruleset": {"name": "ruleset", "label": "Ruleset", "value": "rdfsplus-optimized"},
                "readOnly": {"name": "readOnly", "label": "Read-only", "value": "false"},
                "enableFtsIndex": {
                    "name": "enableFtsIndex",
                    "label": "Enable full-text search (FTS) index",
                    "value": "false",
                },
                "enableLiteralIndex": {"name": "enableLiteralIndex", "label": "Enable literal index", "value": "true"},
                "enableContextIndex": {"name": "enableContextIndex", "label": "Enable context index", "value": "false"},
                "defaultNS": {
                    "name": "defaultNS",
                    "label": "Default namespaces for imports(';' delimited)",
                    "value": "",
                },
                "baseURL": {"name": "baseURL", "label": "Base URL", "value": "http://example.org/owlim#"},
                "logValidationViolations": {
                    "name": "logValidationViolations",
                    "label": "Log validation violations",
                    "value": "false",
                },
                "globalLogValidationExecution": {
                    "name": "globalLogValidationExecution",
                    "label": "Log every execution step of the SHACL validation",
                    "value": "false",
                },
                "entityIdSize": {"name": "entityIdSize", "label": "Entity ID size", "value": "32"},
                "repositoryType": {"name": "repositoryType", "label": "Repository type", "value": "file-repository"},
                "eclipseRdf4jShaclExtensions": {
                    "name": "eclipseRdf4jShaclExtensions",
                    "label": "RDF4J SHACL extensions",
                    "value": "true",
                },
                "validationResultsLimitTotal": {
                    "name": "validationResultsLimitTotal",
                    "label": "Validation results limit total",
                    "value": 1000000,
                },
            },
            "title": "",
            "type": "graphdb",
            "location": "",
        }
        r = requests.post(f"{self.source_graph_api_url}/rest/repositories", json=req)
        logging.info("Repository created with state: %s", r.text)
        if r.status_code > 202:
            if r.text.find("already exists") != -1:
                logging.info("Repository already exists")
                return FlowMessage(output_text="Repository already exists", output_data={})
            else:
                raise Exception(f"Repo creation failed , error from graphdb :{r.text}")

    def step_load_source_graph(self, flow_msg: FlowMessage = None):
        repo_name = self.get_config_item("graphdb_repo_name").value
        file_name = self.get_config_item("import_file_name").value
        import_prefix = self.get_config_item("import_prefix").value
        file_path = Path(file_name)
        if file_path.is_dir():
            pass
        req = {
            "importSettings": {
                "name": file_name,
                "status": "NONE",
                "message": "",
                "context": "",
                "replaceGraphs": [],
                "baseURI": import_prefix,
                "forceSerial": False,
                "type": "file",
                "format": None,
                "data": None,
                "timestamp": 1674660042160,
                "parserSettings": {
                    "preserveBNodeIds": False,
                    "failOnUnknownDataTypes": False,
                    "verifyDataTypeValues": False,
                    "normalizeDataTypeValues": False,
                    "failOnUnknownLanguageTags": False,
                    "verifyLanguageTags": True,
                    "normalizeLanguageTags": False,
                    "stopOnError": True,
                },
                "requestIdHeadersToForward": None,
            },
            "fileNames": [file_name],
        }
        r = requests.post(f"{self.source_graph_api_url}/rest/repositories/{repo_name}/import/server", json=req)
        logging.info(f"File import operation with code : {r.status_code} state: {r.text} ")
        if r.status_code > 202:
            raise Exception(f"The file import failed , error from graphdb :{r.text}")
        return FlowMessage(output_text="File import started successfully")

    def step_wait_for_import_complition(self, flow_msg: FlowMessage = None):
        repo_name = self.get_config_item("graphdb_repo_name").value
        for _ in range(1000):
            time.sleep(2)
            r = requests.get(f"{self.source_graph_api_url}/rest/repositories/{repo_name}/import/server")

            resp = r.json()
            logging.debug(f"Import status : {resp}")
            if resp[0]["status"] == "IMPORTING":
                continue
            elif resp[0]["status"] == "DONE":
                output_text = "File imported successfully . Message :" + resp[0]["message"]
                logging.info(output_text)
                return FlowMessage(output_text=output_text)
            else:
                raise Exception("Import failed with error :" + resp[0]["message"])
        raise Exception("Import operation timed out")

    def step_cleanup(self, flow_msg: FlowMessage = None):
        # TODO : cleanup
        self.categorized_assets = None
        self.categorized_relationships = None



import logging
import os
import shutil
import time
import traceback
from contextlib import asynccontextmanager
from logging import getLogger
from pathlib import Path

from fastapi import FastAPI, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
from prometheus_client import REGISTRY, Counter, make_asgi_app

from cognite import neat
from cognite.neat import constants
from cognite.neat.core import loader, parser, query_generator
from cognite.neat.core.app import NeatApp
from cognite.neat.core.data_classes.config import Config, configure_logging
from cognite.neat.core.loader.config import copy_examples_to_directory
from cognite.neat.core.workflow import WorkflowFullStateReport, utils
from cognite.neat.core.workflow.base import WorkflowDefinition
from cognite.neat.explorer.data_classes.rest import (
    DownloadFromCdfRequest,
    NodesAndEdgesRequest,
    QueryRequest,
    RuleRequest,
    RunWorkflowRequest,
    UploadToCdfRequest,
)
from cognite.neat.explorer.utils.data_mapping import rdf_result_to_api_response
from cognite.neat.explorer.utils.query_templates import query_templates

logger = getLogger(__name__)  # temporary logger before config is loaded
config_path = Path(os.environ.get("NEAT_CONFIG_PATH", "config.yaml"))

if os.environ.get("NEAT_CDF_PROJECT"):
    logger.info("ENV NEAT_CDF_PROJECT is set, loading config from env.")
    config = Config.from_env()
elif (config_path := Path(os.environ.get("NEAT_CONFIG_PATH", "config.yaml"))).exists():
    logger.info(f"Loading config from {config_path.name}.")
    config = Config.from_yaml(config_path)
else:
    logger.error(f"Config file {config_path.name} not found.Exiting.")
    config = Config()
    config.to_yaml(config_path)

if config.load_examples:
    copy_examples_to_directory(config.data_store_path)

configure_logging(config.log_level, config.log_format)
logging.info(f" Starting NEAT version {neat.__version__}")
logging.debug(f" Config: {config.dict(exclude={'cdf_client': {'client_secret': ...}})}")

neat_app = NeatApp(config)


@asynccontextmanager
async def lifespan(app_ref: FastAPI):
    logging.info("Startup FastAPI server")
    neat_app.set_http_server(app_ref)
    neat_app.start()
    yield
    logging.info("FastApi shutdown event")
    neat_app.stop()


app = FastAPI(title="Neat", lifespan=lifespan)


origins = [
    "http://localhost:8000",
    "http://localhost:3000",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# UI cache store
cache_store = {}


counter = Counter("started_workflows", "Description of counter")

prom_app = make_asgi_app()
app.mount("/metrics", prom_app)
app.mount("/static", StaticFiles(directory=constants.UI_PATH), name="static")


@app.get("/")
def read_root():
    return RedirectResponse("/static/index.html")


@app.get("/api/about")
def get_about():
    return {"version": neat.__version__}


@app.get("/api/configs/global")
def get_configs():
    return config.dict()


@app.post("/api/configs/global")
def set_configs(request: Config):
    logging.info(f"Updating global config: {request}")
    global config
    global cache_store
    global neat_app
    config = request
    config.to_yaml(config_path)
    cache_store = {}
    neat_app.stop()
    neat_app.start(config=config)
    return config


@app.get("/api/rules")
def get_rules(
    sheetname: str = "Properties",
    url: str | None = None,
    source_type: str | None = None,
    orient: str = "columns",
    workflow_name: str = "default",
    file_name: str | None = None,
    version: str | None = None,
):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    if not file_name:
        version = workflow.get_config_item("rules.version").value
        file_name = workflow.get_config_item("rules.file").value
    path = Path(config.rules_store_path, file_name)
    src = "local"
    if url:
        path = url

    if path.exists() and not version:
        logging.info(f"Loading rules from {path}")
    elif path.exists() and version:
        hash = utils.get_file_hash(path)
        if hash != version:
            neat_app.cdf_store.load_rules_file_from_cdf(file_name, version)
            src = "cdf"
    else:
        neat_app.cdf_store.load_rules_file_from_cdf(file_name, version)
        src = "cdf"

    tables = loader.rules.excel_file_to_table_by_name(path)
    error_text = ""
    properties = []
    classes = []
    try:
        rules = parser.parse_transformation_rules(tables)
        properties = [
            {
                "class": value.class_name,
                "property": value.property_name,
                "property_description": value.description,
                "property_type": value.expected_value_type,
                "cdf_resource_type": value.cdf_resource_type,
                "cdf_metadata_type": value.resource_type_property,
                "rule_type": value.rule_type,
                "rule": value.rule,
            }
            for value in rules.properties.values()
        ]

        classes = [
            {
                "class": value.class_name,
                "class_description": value.description,
                "cdf_resource_type": value.cdf_resource_type,
                "cdf_parent_resource": value.parent_asset,
            }
            for value in rules.classes.values()
        ]
    except Exception as e:
        error_text = str(e)

    return {
        "properties": properties,
        "classes": classes,
        "file_name": path.name,
        "hash": utils.get_file_hash(path),
        "error_text": error_text,
        "src": src,
    }


@app.get("/api/list-queries")
def list_queries():
    counter.inc()
    return query_templates


def get_data_from_graph(sparq_query: str, graph_name: str = "source", workflow_name: str = "default"):
    total_elapsed_time = 0
    api_result = {"error": ""}
    result = None

    try:
        logging.info(f"Preparing query :{sparq_query} ")
        start_time = time.perf_counter()
        workflow = neat_app.workflow_manager.get_workflow(workflow_name)
        try:
            if not workflow.source_graph or not workflow.solution_graph:
                workflow.step_load_transformation_rules()
                workflow.step_configuring_stores(clean_start=False)
        except Exception as e:
            logging.debug(f"Error while loading transformation rules or stores : {e}")

        if graph_name == "source":
            if workflow.source_graph:
                result = workflow.source_graph.query(sparq_query)
            else:
                logging.info("Source graph is empty , please load the graph first")
                api_result["error"] = "source graph is empty , please load the graph first"
        elif graph_name == "solution":
            if workflow.solution_graph:
                result = workflow.solution_graph.query(sparq_query)
            else:
                logging.info("Solution graph is empty , please load the graph first")
                api_result["error"] = "solution graph is empty , please load the graph first"
        else:
            raise Exception("Unknown graph name")

        stop_time = time.perf_counter()
        elapsed_time_sec_1 = stop_time - start_time
        logging.info(f"Query prepared in {elapsed_time_sec_1 * 1000} ms")

        start_time = time.perf_counter()
        if result:
            api_result = rdf_result_to_api_response(result)
        stop_time = time.perf_counter()
        elapsed_time_sec_2 = stop_time - start_time
        total_elapsed_time = elapsed_time_sec_1 + elapsed_time_sec_2
    except Exception as e:
        logging.error(f"Error while executing query :{e}")
        traceback.print_exc()
        api_result["error"] = str(e)

    api_result["query"] = sparq_query
    api_result["elapsed_time_sec"] = total_elapsed_time
    logging.info(f"Data fetched in {total_elapsed_time * 1000} ms")
    return api_result


@app.post("/api/query")
def query_graph(request: QueryRequest):
    logging.info(f"Querying graph {request.graph_name} with query {request.query}")
    sparq_query = request.query
    return get_data_from_graph(sparq_query, request.graph_name, workflow_name=request.workflow_name)


@app.post("/api/execute-rule")
def execute_rule(request: RuleRequest):
    logging.info(
        f"Executing rule type: { request.rule_type } rule : {request.rule} , workflow : {request.workflow_name} , graph : {request.graph_name}"
    )
    # TODO : add support for other graphs
    workflow = neat_app.workflow_manager.get_workflow(request.workflow_name)
    if not workflow.source_graph or not workflow.solution_graph:
        workflow.step_load_transformation_rules()
        workflow.step_configuring_stores()
    if request.graph_name == "source":
        if not workflow.source_graph:
            logging.info("Source graph is empty , please load the graph first")
            return {"error": "source graph is empty , please load the graph first"}
        graph = workflow.source_graph
    elif request.graph_name == "solution":
        if not workflow.solution_graph:
            logging.info("Solution graph is empty , please load the graph first")
            return {"error": "solution graph is empty , please load the graph first"}
        graph = workflow.solution_graph
    else:
        raise Exception("Unknown graph name")

    if request.rule_type == "rdfpath":
        start_time = time.perf_counter()
        sparq_query = query_generator.build_sparql_query(
            graph, request.rule, prefixes=workflow.transformation_rules.prefixes
        )
    else:
        logging.error("Unknown rule type")
        return {"error": "Unknown rule type"}
    stop_time = time.perf_counter()
    elapsed_time_sec_1 = stop_time - start_time
    logging.info(f"Computed query : {sparq_query} in {elapsed_time_sec_1 * 1000} ms")

    api_result = get_data_from_graph(sparq_query, request.graph_name, workflow_name=request.workflow_name)
    api_result["elapsed_time_sec"] += elapsed_time_sec_1
    return api_result


@app.get("/api/object-properties")
def get_object_properties(reference: str, graph_name: str = "source", workflow_name: str = "default"):
    logging.info(f"Querying object-properties from {graph_name} :")
    query = f"SELECT ?property ?value WHERE {{ <{reference}> ?property ?value }}"
    return get_data_from_graph(query, graph_name, workflow_name=workflow_name)


@app.get("/api/search")
def search(
    search_str: str, graph_name: str = "source", search_type: str = "value_exact_match", workflow_name: str = "default"
):
    logging.info(f"Search from {graph_name} :")
    if search_type == "reference":
        query = (
            f"SELECT ?class ?reference WHERE {{ {{?reference ?class <{search_str}> . }} "
            f"UNION {{ <{search_str}> ?class ?reference }} }} limit 10000"
        )

    elif search_type == "value_exact_match":
        query = (
            f"select ?object_ref ?type ?property ?value where "
            f"{{ ?object_ref ?property ?value . ?object_ref rdf:type ?type . "
            f'FILTER(?value="{search_str}") }} limit 10000'
        )
    elif search_type == "value_regexp":
        query = (
            f"select ?object_ref ?type ?property ?value where "
            f"{{ ?object_ref ?property ?value . ?object_ref rdf:type ?type .  "
            f'FILTER regex(?value, "{search_str}","i") }} limit 10000'
        )
    return get_data_from_graph(query, graph_name, workflow_name=workflow_name)


@app.get("/api/get-classes")
def get_classes(graph_name: str = "source", workflow_name: str = "default", cache: bool = True):
    logging.info(f"Querying graph classes from graph name : {graph_name}, cache : {cache}")
    cache_key = f"classes_{graph_name}"
    if cache_key in cache_store and cache:
        return cache_store[cache_key]
    query = "SELECT ?class (count(?s) as ?instances ) WHERE { ?s rdf:type ?class . } group by ?class order by DESC(?instances)"
    api_result = get_data_from_graph(query, graph_name, workflow_name)
    cache_store["get_classes"] = api_result
    return api_result


@app.post("/api/get-nodes-and-edges")
def get_nodes_and_edges(request: NodesAndEdgesRequest):
    logging.info("Querying graph nodes and edges :")
    if "get_nodes_and_edges" in cache_store and request.cache:
        return cache_store["get_nodes_and_edges"]
    nodes_filter = ""
    edges_dst_filter = ""
    edges_src_filter = ""

    if len(request.node_class_filter) > 0:
        nodes_filter = "VALUES ?node_class { " + " ".join([f"<{x}>" for x in request.node_class_filter]) + " }"

    node_name_property = request.node_name_property or "cim:IdentifiedObject.name"
    nodes_query = f"SELECT DISTINCT ?node_id ?node_class ?node_name WHERE \
    {{ {nodes_filter} \
    ?node_id {node_name_property} ?node_name . ?node_id rdf:type ?node_class }} "
    if len(request.src_edge_filter) > 0:
        edges_src_filter = "VALUES ?src_object_class { " + " ".join([f"<{x}>" for x in request.src_edge_filter]) + " }"
    if len(request.dst_edge_filter) > 0:
        edges_dst_filter = "VALUES ?dst_object_class { " + " ".join([f"<{x}>" for x in request.dst_edge_filter]) + " }"

    edges_query = f"SELECT  ?src_object_ref ?conn ?dst_object_ref \
        WHERE {{ \
        {edges_src_filter} \
        {edges_dst_filter} \
        ?src_object_ref ?conn  ?dst_object_ref . \
        ?src_object_ref  rdf:type ?src_object_class . \
        ?dst_object_ref  rdf:type ?dst_object_class .\
        }} "
    logging.info(f"Nodes query : {nodes_query}")
    logging.info(f"Edges query : {edges_query}")
    nodes_result = get_data_from_graph(nodes_query, request.graph_name, request.workflow_name)
    edges_result = get_data_from_graph(edges_query, request.graph_name, request.workflow_name)

    merged_result = {
        "nodes": nodes_result["rows"],
        "edges": edges_result["rows"],
        "error": "",
        "elapsed_time_sec": nodes_result["elapsed_time_sec"] + edges_result["elapsed_time_sec"],
        "query": nodes_query + edges_query,
    }

    if request.cache:
        cache_store["get_nodes_and_edges"] = merged_result
    return merged_result


@app.post("/api/workflow/start")
def start_workflow(request: RunWorkflowRequest):
    logging.info("Starting workflow endpoint")
    workflow = neat_app.workflow_manager.get_workflow(request.name)
    result = workflow.start(sync=request.sync)
    return {"result": result}


@app.get("/api/workflow/stats/{workflow_name}", response_model=WorkflowFullStateReport)
def get_workflow_stats(
    workflow_name: str,
) -> WorkflowFullStateReport:
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    return workflow.get_state()


@app.get("/api/workflow/workflows")
def get_workflows():
    return {"workflows": neat_app.workflow_manager.get_list_of_workflows()}


@app.get("/api/workflow/executions")
def get_list_of_workflow_executions():
    return {"executions": neat_app.cdf_store.get_list_of_workflow_executions_from_cdf()}


@app.get("/api/workflow/detailed-execution-report/{execution_id}")
def get_detailed_execution(execution_id: str):
    return {"report": neat_app.cdf_store.get_detailed_workflow_execution_report_from_cdf(execution_id)}


@app.post("/api/workflow/reload-workflows")
def reload_workflows():
    neat_app.workflow_manager.load_workflows_from_storage_v2()
    neat_app.triggers_manager.reload_all_triggers()
    return {"result": "ok", "workflows": neat_app.workflow_manager.get_list_of_workflows()}


@app.get("/api/workflow/workflow-definition/{workflow_name}")
def get_workflow_definition(workflow_name: str):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    return {"definition": workflow.get_workflow_definition()}


@app.post("/api/workflow/workflow-definition/{workflow_name}")
def update_workflow_definition(workflow_name: str, request: WorkflowDefinition):
    neat_app.workflow_manager.update_workflow(workflow_name, request)
    neat_app.workflow_manager.save_workflow_to_storage(workflow_name)
    return {"result": "ok"}


@app.post("/api/workflow/upload-wf-to-cdf/{workflow_name}")
def upload_workflow_to_cdf(workflow_name: str, request: UploadToCdfRequest):
    neat_app.cdf_store.save_workflow_to_cdf(
        workflow_name, changed_by=request.author, comments=request.comments, tag=request.tag
    )
    return {"result": "ok"}


@app.post("/api/workflow/upload-rules-cdf/{workflow_name}")
def upload_rules_to_cdf(workflow_name: str, request: UploadToCdfRequest):
    file_path = Path(config.rules_store_path, request.file_name)
    neat_app.cdf_store.save_resource_to_cdf(
        workflow_name, "neat-wf-rules", file_path, changed_by=request.author, comments=request.comments
    )
    return {"result": "ok"}


@app.post("/api/workflow/download-wf-from-cdf")
def download_wf_from_cdf(request: DownloadFromCdfRequest):
    neat_app.cdf_store.load_workflows_from_cdf(request.file_name, request.version)
    return {"result": "ok"}


@app.post("/api/workflow/download-rules-from-cdf")
def download_rules_to_cdf(request: DownloadFromCdfRequest):
    neat_app.cdf_store.load_rules_file_from_cdf(request.file_name, request.version)
    return {"file_name": request.file_name, "hash": request.version}


@app.get("/api/workflow/pre-cdf-assets/{workflow_name}")
def get_pre_cdf_assets(workflow_name: str):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return {"assets": []}
    return {"assets": workflow.categorized_assets}


@app.get("/api/metrics")
def get_metrics():
    metrics = REGISTRY.collect()
    return {"prom_metrics": metrics}


@app.get("/api/cdf/neat-resources")
def get_neat_resources(resource_type: str = None):
    result = neat_app.cdf_store.get_list_of_resources_from_cdf(resource_type=resource_type)
    logging.debug(f"Got {len(result)} resources")
    return {"result": result}


@app.post("/api/cdf/init-neat-resources")
def init_neat_cdf_resources(resource_type: str = None):
    neat_app.cdf_store.init_cdf_resources(resource_type=resource_type)
    return {"result": "ok"}


@app.post("/api/file/upload/{workflow_name}/{file_type}")
async def file_upload_handler(files: list[UploadFile], workflow_name: str, file_type: str):
    # get directory path
    upload_dir = config.rules_store_path
    file_name = ""
    file_version = ""
    for file in files:
        logging.info(f"Uploading file : {file.filename} , workflow : {workflow_name}")
        # save file to disk
        full_path = os.path.join(upload_dir, file.filename)
        with open(full_path, "wb") as buffer:
            shutil.copyfileobj(file.file, buffer)
        file_name = file.filename
        file_version = utils.get_file_hash(full_path)
        break  # only one file is supported for now

    return {"file_name": file_name, "hash": file_version}



from collections import OrderedDict
from datetime import datetime, timezone

import pandas as pd
from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import CredentialProvider, OAuthClientCredentials, OAuthInteractive
from rdflib.term import URIRef

from cognite.neat.core.data_classes.config import InteractiveClient, ServiceClient


def get_cognite_client_from_config(config: ServiceClient) -> CogniteClient:
    credentials = OAuthClientCredentials(
        token_url=config.token_url, client_id=config.client_id, client_secret=config.client_secret, scopes=config.scopes
    )

    return _get_cognite_client(config, credentials)


def get_cognite_client_interactive(config: InteractiveClient) -> CogniteClient:
    credentials = OAuthInteractive(
        authority_url=config.authority_url,
        client_id=config.client_id,
        scopes=config.scopes,
        redirect_port=config.redirect_port,
    )
    return _get_cognite_client(config, credentials)


def _get_cognite_client(config: ClientConfig, credentials: CredentialProvider) -> CogniteClient:
    return CogniteClient(
        ClientConfig(
            client_name=config.client_name,
            base_url=config.base_url,
            project=config.project,
            credentials=credentials,
            timeout=60,
            max_workers=3,
            debug=False,
        )
    )


def remove_namespace(URI: URIRef, special_separator: str = "#_") -> str:
    """Removes namespace from URI

    Parameters
    ----------
    URI : URIRef
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers Statnett use case

    Returns
    -------
    str
        Entity id without namespace
    """

    return URI.split(special_separator if special_separator in URI else ("#" if "#" in URI else "/"))[-1]


def _traverse(hierarchy: dict, graph: dict, names: str) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def get_generation_order(
    class_linkage: pd.DataFrame, parent_col: str = "source_class", child_col: str = "target_class"
) -> dict:
    parent_child_list = class_linkage[[parent_col, child_col]].values.tolist()
    # Build a directed graph and a list of all names that have no parent
    graph = {name: set() for tup in parent_child_list for name in tup}
    has_parent = {name: False for tup in parent_child_list for name in tup}
    for parent, child in parent_child_list:
        graph[parent].add(child)
        has_parent[child] = True

    # All names that have absolutely no parent:
    roots = [name for name, parents in has_parent.items() if not parents]

    return _traverse({}, graph, roots)


def prettify_generation_order(generation_order: dict, depth: dict = None, start=-1) -> dict:
    """Prettifies generation order dictionary for easier consumption."""
    depth = depth or {}
    for key, value in generation_order.items():
        depth[key] = start + 1
        if isinstance(value, dict):
            prettify_generation_order(value, depth, start=start + 1)
    return OrderedDict(sorted(depth.items(), key=lambda item: item[1]))


def epoch_now_ms():
    return int((datetime.now(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds() * 1000)


def chunker(sequence, chunk_size):
    return [sequence[i : i + chunk_size] for i in range(0, len(sequence), chunk_size)]


def datetime_utc_now():
    return datetime.now(timezone.utc)



import json
import logging
import threading
import time
import traceback

import yaml
from cognite.client import ClientConfig as CogniteClientConfig
from cognite.client import CogniteClient
from prometheus_client import Gauge

from cognite.neat.core.data_classes.config import ClientConfig, Config
from cognite.neat.core.workflow import cdf_store
from cognite.neat.core.workflow.model import (
    FlowMessage,
    StepExecutionStatus,
    StepType,
    WorkflowConfigItem,
    WorkflowDefinition,
    WorkflowFullStateReport,
    WorkflowState,
    WorkflowStepDefinition,
    WorkflowStepEvent,
    WorkflowStepsGroup,
)
from cognite.neat.core.workflow.tasks import WorkflowTaskBuilder

from . import utils

summary_metrics = Gauge("neat_workflow_summary_metrics", "Workflow execution summary metrics", ["wf_name", "name"])
steps_metrics = Gauge("neat_workflow_steps_metrics", "Workflow step level metrics", ["wf_name", "step_name", "name"])
timing_metrics = Gauge("neat_workflow_timing_metrics", "Workflow timing metrics", ["wf_name", "component", "name"])


class BaseWorkflow:
    def __init__(
        self,
        name: str,
        client: CogniteClient,
        workflow_steps: list[WorkflowStepDefinition] = None,
        default_dataset_id: int = None,
    ):
        self.name = name
        self.module_name = self.__class__.__module__
        self.cdf_client = client
        self.cdf_client_config: CogniteClientConfig = client.config
        self.default_dataset_id = default_dataset_id
        self.state = WorkflowState.CREATED
        self.run_id = ""
        self.last_error = ""
        self.elapsed_time = 0
        self.start_time = None
        self.end_time = None
        self.execution_log: list[WorkflowStepEvent] = []
        self.workflow_steps: list[WorkflowStepDefinition] = workflow_steps
        self.workflow_step_groups: list[WorkflowStepsGroup] = []
        self.configs: list[WorkflowConfigItem] = []
        self.flow_message: FlowMessage = None
        self.task_builder: WorkflowTaskBuilder = None
        self.rules_storage_path = None
        self.cdf_store = (
            cdf_store.CdfStore(self.cdf_client, data_set_id=self.default_dataset_id)
            if self.default_dataset_id
            else None
        )

    def start(self, sync=False, **kwargs) -> FlowMessage | None:
        """Starts workflow execution.sync=True will block until workflow is completed and return last workflow flow message,
        sync=False will start workflow in a separate thread and return None"""
        self.execution_log = []
        if sync:
            return self._run_workflow(**kwargs)

        self.thread = threading.Thread(target=self._run_workflow, kwargs=kwargs)
        self.thread.start()
        return None

    def _run_workflow(self, **kwargs) -> FlowMessage | None:
        """Run workflow and return last workflow flow message"""
        summary_metrics.labels(wf_name=self.name, name="steps_count").set(len(self.workflow_steps))
        if self.state not in [WorkflowState.CREATED, WorkflowState.COMPLETED, WorkflowState.FAILED]:
            logging.error(f"Workflow {self.name} is already running")
            return None
        logging.info("Starting workflow")

        if flow_message := kwargs.get("flow_message"):
            self.flow_message = flow_message

        self.state = WorkflowState.RUNNING
        self.start_time = time.time()
        self.end_time = None
        self.run_id = utils.generate_run_id()
        start_time = time.perf_counter()
        self.report_workflow_execution()
        try:
            start_step_id = kwargs.get("start_step_id")
            self.run_workflow_steps(start_step_id=start_step_id)
            if self.state == WorkflowState.RUNNING:
                self.state = WorkflowState.COMPLETED
            summary_metrics.labels(wf_name=self.name, name="wf_completed_counter").inc()
        except Exception:
            trace = traceback.format_exc()
            self.last_error = str(trace)
            self.state = WorkflowState.FAILED
            logging.error(f"Workflow failed with error {trace}")
            summary_metrics.labels(wf_name=self.name, name="wf_failed_counter").inc()

        self.elapsed_time = time.perf_counter() - start_time
        self.end_time = time.time()
        timing_metrics.labels(wf_name=self.name, component="workflow", name="workflow").set(self.elapsed_time)
        logging.info(f"Workflow completed in {self.elapsed_time} seconds")
        self.report_workflow_execution()
        return self.flow_message

    def get_transition_step(self, transitions: list[str]) -> list[str]:
        return [stp for stp in self.workflow_steps if stp.id in transitions and stp.enabled] if transitions else []

    def run_workflow_steps(self, start_step_id: str = None) -> str:
        if start_step_id is None:
            trigger_steps = list(filter(lambda x: x.trigger, self.workflow_steps))
        else:
            trigger_steps = list(filter(lambda x: x.id == start_step_id, self.workflow_steps))

        if not trigger_steps:
            logging.error(f"Workflow {self.name} has no trigger steps")
            return "Workflow has no trigger steps"

        step: WorkflowStepDefinition = trigger_steps[0]

        transition_steps = self.get_transition_step(step.transition_to)

        if len(transition_steps) == 0:
            logging.error(f"Workflow {self.name} has no transition steps from step {step.id}")
            return "Workflow has no transition steps"

        step = transition_steps[0]

        for _ in range(
            1000
        ):  # Max 1000 steps in a workflow is a reasonable limit and protection against infinite loops
            new_flow_msg = None
            if step.enabled:
                new_flow_msg = self.run_step(step)
            else:
                logging.info(f"Skipping step workflow step {step.id}")
                self.execution_log.append(
                    WorkflowStepEvent(
                        id=step.id,
                        group_id=step.group_id,
                        state=StepExecutionStatus.SKIPPED,
                        elapsed_time=0,
                        timestamp=utils.get_iso8601_timestamp_now_unaware(),
                        error="",
                    )
                )

            steps = self.get_transition_step(step.transition_to)

            if new_flow_msg:
                # If the step returned a new flow message, use it to mutate execution flow
                if new_flow_msg.next_step_ids:
                    steps = self.get_transition_step(new_flow_msg.next_step_ids)
                self.flow_message = new_flow_msg

            if len(steps) == 0:
                break
            step = steps[0]
            logging.debug(f"Transitioning to step {step.id}")

        return "ok"

    def configure(self, config: Config):
        raise NotImplementedError()

    def run_step(self, step: WorkflowStepDefinition) -> FlowMessage | None:
        step_name = step.id
        group_id = step.group_id

        steps_metrics.labels(wf_name=self.name, step_name=step_name, name="step_started_counter").inc()
        flow_message = self.flow_message
        if self.state != WorkflowState.RUNNING:
            logging.error(f"Workflow {self.name} is not running , step {step_name} is skipped")
            return

        logging.info(f"Running step {step_name}")
        self.current_step = step_name
        start_time = time.perf_counter()
        stop_time = start_time
        step_execution_status = StepExecutionStatus.STARTED
        self.execution_log.append(
            WorkflowStepEvent(
                id=step_name,
                group_id=group_id,
                state=step_execution_status,
                elapsed_time=0,
                timestamp=utils.get_iso8601_timestamp_now_unaware(),
                error="",
            )
        )
        output_text = ""
        error_text = ""
        new_flow_message = None
        try:
            if step.stype == StepType.PYSTEP:
                if step.method and hasattr(self, step.method):
                    method = getattr(self, step.method)
                else:
                    method_name = f"step_{step.id}"
                    if hasattr(self, method_name):
                        method = getattr(self, method_name)
                    else:
                        logging.error(f"Workflow step {step.id} has no method {method_name}")
                        raise Exception(f"Workflow step {step.id} has no method {method_name}")
                new_flow_message = method(flow_message)
            elif step.stype == StepType.START_WORKFLOW_TASK_STEP:
                if self.task_builder:
                    sync_str = step.params.get("sync", "false")
                    sync = sync_str.lower() == "true" or sync_str == "1"
                    new_flow_message = self.task_builder.start_workflow_task(
                        workflow_name=step.params.get("workflow_name", ""), sync=sync, flow_message=self.flow_message
                    )
                else:
                    logging.error(f"Workflow step {step.id} has no task builder")
                    raise Exception(f"Workflow step {step.id} has no task builder")
            else:
                logging.error(f"Workflow step {step.id} has unsupported step type {step.stype}")

            stop_time = time.perf_counter()
            elapsed_time = stop_time - start_time
            logging.info(f"Step {step_name} completed in {elapsed_time} seconds")
            step_execution_status = StepExecutionStatus.SUCCESS
            if new_flow_message:
                error_text = new_flow_message.error_text
                output_text = new_flow_message.output_text
            steps_metrics.labels(wf_name=self.name, step_name=step_name, name="completed_counter").inc()
            timing_metrics.labels(wf_name=self.name, component="step", name=step_name).set(elapsed_time)
        except Exception:
            self.state = WorkflowState.FAILED
            step_execution_status = StepExecutionStatus.FAILED
            trace = traceback.format_exc()
            elapsed_time = stop_time - start_time
            logging.error(f"Step {step_name} failed with error : {trace}")
            error_text = str(trace)
            traceback.print_exc()
            steps_metrics.labels(wf_name=self.name, step_name=step_name, name="failed_counter").inc()

        self.execution_log.append(
            WorkflowStepEvent(
                id=step_name,
                group_id=group_id,
                state=step_execution_status,
                elapsed_time=round(elapsed_time, 3),
                timestamp=utils.get_iso8601_timestamp_now_unaware(),
                error=error_text,
                output_text=output_text,
                data=new_flow_message.payload if new_flow_message else None,
            )
        )
        if new_flow_message:
            self.flow_message = new_flow_message

        self.report_step_execution()
        return new_flow_message

    def report_step_execution(self):
        pass

    def report_workflow_execution(self):
        reporting_type = "all_enabled"
        if config_item := self.get_config_item("system.execution_reporting_type"):
            reporting_type = config_item.value
        if reporting_type == "all_disabled":
            return
        if self.cdf_store:
            logging.debug("Reporting workflow execution to CDF")
            try:
                self.cdf_store.report_workflow_execution_to_cdf(self.get_state())
            except Exception:
                logging.error("Failed to report workflow execution to CDF")
                traceback.print_exc()

    def get_state(self):
        return WorkflowFullStateReport(
            workflow_name=self.name,
            state=self.state,
            run_id=self.run_id,
            start_time=self.start_time,
            end_time=self.end_time,
            elapsed_time=round(self.elapsed_time, 3),
            last_error=self.last_error,
            execution_log=self.execution_log,
        )

    def set_cognite_client(self, client: CogniteClient):
        self.cdf_client = client

    def get_workflow_definition(self):
        return WorkflowDefinition(
            name=self.name,
            steps=self.workflow_steps,
            groups=self.workflow_step_groups,
            configs=self.configs,
        )

    def add_step(self, step: WorkflowStepDefinition):
        self.workflow_steps.append(step)

    def add_group(self, group: WorkflowStepsGroup):
        self.workflow_step_groups.append(group)

    def serialize_workflow(self, output_format: str = "json", custom_implementation_module: str = None) -> str:
        workflow_definitions = WorkflowDefinition(
            name=self.name,
            steps=self.workflow_steps,
            groups=self.workflow_step_groups,
            configs=self.configs,
            implementation_module=custom_implementation_module,
        )
        if output_format == "json":
            return json.dumps(workflow_definitions.dict(), indent=4)
        elif output_format == "yaml":
            return yaml.dump(workflow_definitions.dict(), indent=4)

    @classmethod
    def deserialize_metadata(cls, json_string: str, output_format: str = "json") -> WorkflowDefinition:
        if output_format == "json":
            workflow_definitions = WorkflowDefinition.parse_raw(json_string)
        elif output_format == "yaml":
            workflow_definitions = WorkflowDefinition.parse_obj(yaml.load(json_string, Loader=yaml.Loader))
        else:
            raise NotImplementedError(f"Output format {output_format} is not supported.")
        return workflow_definitions

    def set_metadata(self, metadata: WorkflowDefinition):
        self.workflow_steps = metadata.steps
        self.workflow_step_groups = metadata.groups
        self.configs = metadata.configs

    def set_storage_path(self, storage_type: str, storage_path: str):
        if storage_type == "transformation_rules":
            self.rules_storage_path = storage_path

    def set_task_builder(self, task_builder: WorkflowTaskBuilder):
        self.task_builder = task_builder

    def get_config_item(self, config_name: str) -> WorkflowConfigItem:
        return next((item for item in self.configs if item.name == config_name), None)

    def get_config_item_value(self, config_name: str, default_value=None) -> str | None:
        return config.value if (config := self.get_config_item(config_name)) else default_value

    def set_default_dataset_id(self, default_dataset_id: int):
        self.default_dataset_id = default_dataset_id
        self.cdf_store = (
            cdf_store.CdfStore(self.cdf_client, data_set_id=self.default_dataset_id)
            if self.default_dataset_id
            else None
        )

    def set_cdf_client_config(self, cdf_client_config: ClientConfig):
        """Set the CogniteClient configuration to be used by the workflow to create new CogniteClient instances."""
        self.cdf_client_config = cdf_client_config

    def get_new_cdf_client(self):
        """Get a new CogniteClient instance with the same configuration as the one used by the workflow .
          Should be used from workflow steps to avoid sharing the same client instance between steps or reset reference to old client instance.
        Returns: CogniteClient
        """
        return CogniteClient(self.cdf_client_config)



import logging

from cognite.client import CogniteClient
from fastapi import FastAPI

from cognite.neat.core.data_classes.config import Config
from cognite.neat.core.utils import get_cognite_client_from_config
from cognite.neat.core.workflow.cdf_store import CdfStore
from cognite.neat.core.workflow.manager import WorkflowManager
from cognite.neat.core.workflow.triggers import TriggerManager


class NeatApp:
    def __init__(self, config: Config, cdf_client: CogniteClient = None):
        self.config = config
        self.cdf_client: CogniteClient = None
        self.cdf_store: CdfStore = None
        self.workflow_manager: WorkflowManager = None
        self.triggers_manager: TriggerManager = None
        self.fast_api_app: FastAPI = None
        self.cdf_client = cdf_client

    def set_http_server(self, fast_api_app: FastAPI):
        """Set the http server to be used by the triggers manager"""
        self.fast_api_app = fast_api_app

    def start(self, config: Config = None):
        logging.info("Starting NeatApp")
        if config:
            self.config = config
        logging.info("Initializing global objects")
        if not self.cdf_client:
            self.cdf_client = get_cognite_client_from_config(self.config.cdf_client)
        self.cdf_store = CdfStore(
            self.cdf_client,
            self.config.cdf_default_dataset_id,
            workflows_storage_path=self.config.workflows_store_path,
            rules_storage_path=self.config.rules_store_path,
        )

        # Automatically downloading workflows from CDF if enabled in config
        if self.config.workflow_downloader_filter:
            self.cdf_store.load_workflows_from_cfg_by_filter(self.config.workflow_downloader_filter)

        self.workflow_manager = WorkflowManager(
            self.cdf_client,
            self.config.workflows_store_type,
            self.config.workflows_store_path,
            self.config.rules_store_path,
            self.config.cdf_default_dataset_id,
        )
        self.workflow_manager.load_workflows_from_storage_v2()
        self.triggers_manager = TriggerManager(workflow_manager=self.workflow_manager)
        if self.fast_api_app:
            self.triggers_manager.start_http_listeners(self.fast_api_app)
        self.triggers_manager.start_time_schedulers()
        logging.info("NeatApp started")

    def stop(self):
        logging.info("Stopping NeatApp")
        self.triggers_manager.stop_scheduler_main_loop()
        self.cdf_client = None
        self.cdf_store = None
        self.workflow_manager = None
        self.triggers_manager = None
        logging.info("NeatApp stopped")



from pathlib import Path

from cognite import neat

PACKAGE_DIRECTORY = Path(neat.__file__).parent
EXAMPLES_DIRECTORY = PACKAGE_DIRECTORY / "examples"
EXAMPLE_RULES = EXAMPLES_DIRECTORY / "rules"
EXAMPLE_SOURCE_GRAPHS = EXAMPLES_DIRECTORY / "source_graphs"
EXAMPLE_WORKFLOWS = EXAMPLES_DIRECTORY / "workflows"

TNT_TRANSFORMATION_RULES = EXAMPLE_RULES / "Rules-Nordic44-to-TNT.xlsx"
SIMPLE_TRANSFORMATION_RULES = EXAMPLE_RULES / "sheet2cdf-transformation-rules.xlsx"
NORDIC44_KNOWLEDGE_GRAPH = EXAMPLE_SOURCE_GRAPHS / "Knowledge-Graph-Nordic44.xml"
UI_PATH = PACKAGE_DIRECTORY / "explorer-ui" / "neat-app" / "build"



import logging
import os
from enum import StrEnum
from pathlib import Path
from typing import Literal, Self

import yaml
from pydantic import BaseModel, Field, validator
from yaml import safe_load

LOG_FORMAT = "%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def configure_logging(level: str = "DEBUG", log_format: str = LOG_FORMAT):
    """Configure logging based on config."""
    logging.basicConfig(format=log_format, level=logging.getLevelName(level), datefmt=LOG_DATE_FORMAT)


class RdfStoreType(StrEnum):
    """RDF Store type"""

    MEMORY = "memory"
    FILE = "file"
    GRAPHDB = "graphdb"
    SPARQL = "sparql"
    OXIGRAPH = "oxigraph"


class RulesStoreType(StrEnum):
    """Rules Store type"""

    CDF = "cdf"
    FILE = "file"
    URL = "url"
    GOOGLE_SHEET = "google_sheet"


class WorkflowsStoreType(StrEnum):
    """Workflows Store type"""

    CDF = "cdf"
    FILE = "file"
    URL = "url"


class ClientConfig(BaseModel):
    project: str = "dev"
    client_id: str = "neat"
    client_name: str = "neat"
    base_url: str = "https://api.cognitedata.com"
    scopes: list[str] = ["project:read", "project:write"]

    @validator("scopes", pre=True)
    def string_to_list(cls, value):
        return [value] if isinstance(value, str) else value


class InteractiveClient(ClientConfig):
    authority_url: str
    redirect_port: int = 53_000


class ServiceClient(ClientConfig):
    token_url: str = "https://login.microsoftonline.com/common/oauth2/token"
    client_secret: str = "secret"


class RdfStoreConfig(BaseModel):
    type: RdfStoreType
    file_path: str = None
    query_url: str = None
    update_url: str = None
    user: str = None
    password: str = None
    api_root_url: str = "http://localhost:7200"


class Config(BaseModel):
    workflows_store_type: RulesStoreType = WorkflowsStoreType.FILE
    data_store_path: Path = Field(default_factory=lambda: Path.cwd() / "data")

    workflow_downloader_filter: list[str] = Field(
        description="List of workflow names+tags to filter on when downloading workflows from CDF. "
        "Example name:workflow_name=version,tag:tag_name",
        default=None,
    )

    cdf_client: InteractiveClient | ServiceClient = ServiceClient()
    cdf_default_dataset_id: int = 0
    load_examples: bool = True

    log_level: Literal["ERROR", "WARNING", "INFO", "DEBUG"] = "INFO"
    log_format: str = LOG_FORMAT
    download_workflows_from_cdf: bool = Field(
        default=False, description="Downloads all workflows from CDF automatically and stores them locally"
    )

    stop_on_error: bool = False

    @property
    def rules_store_path(self) -> Path:
        return self.data_store_path / "rules"

    @property
    def workflows_store_path(self) -> Path:
        return self.data_store_path / "workflows"

    @property
    def source_graph_path(self) -> Path:
        return self.data_store_path / "source_graphs"

    @classmethod
    def from_yaml(cls, filepath: Path) -> Self:
        return cls(**safe_load(filepath.read_text()))

    def to_yaml(self, filepath: Path):
        with filepath.open("w") as f:
            yaml.dump(self.dict(), f)

    @classmethod
    def from_env(cls) -> Self:
        cdf_config = ServiceClient(
            project=os.environ.get("NEAT_CDF_PROJECT"),
            client_name=os.environ.get("NEAT_CDF_CLIENT_NAME"),
            client_id=os.environ.get("NEAT_CDF_CLIENT_ID"),
            client_secret=os.environ.get("NEAT_CDF_CLIENT_SECRET"),
            base_url=os.environ.get("NEAT_CDF_BASE_URL"),
            token_url=os.environ.get("NEAT_CDF_TOKEN_URL"),
            scopes=[os.environ.get("NEAT_CDF_SCOPES")],
        )

        if workflow_downloader_filter := os.environ.get("NEAT_WORKFLOW_DOWNLOADER_FILTER", None):
            workflow_downloader_filter = workflow_downloader_filter.split(",")

        return cls(
            cdf_client=cdf_config,
            workflows_store_type=os.environ.get("NEAT_WORKFLOWS_STORE_TYPE", WorkflowsStoreType.FILE),
            workflows_store_path=os.environ.get("NEAT_DATA_PATH", "/app/data"),
            cdf_default_dataset_id=os.environ.get("NEAT_CDF_DEFAULT_DATASET_ID", 6476640149881990),
            log_level=os.environ.get("NEAT_LOG_LEVEL", "INFO"),
            workflow_downloader_filter=workflow_downloader_filter,
        )


EXCLUDE_PATHS = [
    "root['labels']",
    "root['metadata']['create_time']",
    "root['metadata']['start_time']",
    "root['metadata']['update_time']",
    "root['metadata']['end_time']",
    "root['metadata']['resurrection_time']",  # need to account for assets that are brought back to life
]



import logging
import re
import warnings

from graphql import GraphQLError, GraphQLField, GraphQLList, GraphQLNonNull, GraphQLObjectType, GraphQLSchema
from graphql import assert_name as assert_graphql_name
from graphql import print_schema

from cognite.neat.core.data_classes.transformation_rules import DATA_TYPE_MAPPING, Property, TransformationRules


def get_invalid_names(entity_names: set) -> set:
    """Returns a set of invalid entity names"""
    invalid_names = set()
    for entity_name in entity_names:
        try:
            assert_graphql_name(entity_name)
        except GraphQLError:
            invalid_names.add(entity_name)
    return invalid_names


def repair_name(name: str, entity_type: str, fix_casing: bool = False) -> str:
    """
    Repairs an entity name to conform to GraphQL naming convention
    >>> repair_name("wind-speed", "property")
    'windspeed'
    >>> repair_name("Wind.Speed", "property", True)
    'windSpeed'
    >>> repair_name("windSpeed", "class", True)
    'WindSpeed'
    >>> repair_name("22windSpeed", "class")
    '_22windSpeed'
    """

    # Remove any non GraphQL compliant characters
    repaired_string = re.sub(r"[^_a-zA-Z0-9/_]", "", name)

    # Name must start with a letter or underscore
    if repaired_string[0].isdigit():
        repaired_string = f"_{repaired_string}"

    if not fix_casing:
        return repaired_string
    # Property names must be camelCase
    if entity_type == "property" and repaired_string[0].isupper():
        return repaired_string[0].lower() + repaired_string[1:]
    # Class names must be PascalCase
    elif entity_type == "class" and repaired_string[0].islower():
        return repaired_string[0].upper() + repaired_string[1:]
    else:
        return repaired_string


def _remove_query_type(schema_string: str) -> str:
    """Removes unnecessary Query types to conform to Cognite's GraphQL API"""
    lines = schema_string.split("\n")

    for _i, line in enumerate(lines):
        if "}" in line:
            break

    return "\n".join(lines[_i + 2 :])


def _get_graphql_schema_string(schema: GraphQLSchema) -> str:
    return _remove_query_type(print_schema(schema))


def rules2graphql_schema(
    transformation_rules: TransformationRules,
    stop_on_exception: bool = False,
    fix_casing: bool = False,
) -> str:
    """Generates a GraphQL schema from an instance of TransformationRules

    Parameters
    ----------
    transformation_rules : TransformationRules
        TransformationRules object
    stop_on_exception : bool, optional
        Stop on any exception, by default False
    fix_casing : bool, optional
        Whether to attempt to fix casing of entity names, by default False

    Returns
    -------
    str
        GraphQL schema string
    """
    invalid_names: set = get_invalid_names(transformation_rules.get_entity_names())
    data_model_issues: set = transformation_rules.check_data_model_definitions()

    if invalid_names and stop_on_exception:
        msg = "Entity names must only contain [_a-zA-Z0-9] characters and can start only with [_a-zA-Z]"
        logging.error(f"{msg}, following entities {invalid_names} do not follow these rules!")
        raise GraphQLError(f"{msg}, following entities {invalid_names} do not follow these rules!")
    elif invalid_names and not stop_on_exception:
        msg = "Entity names must only contain [_a-zA-Z0-9] characters and can start only with [_a-zA-Z]"
        logging.warn(
            f"{msg}, following entities {invalid_names} do not follow these rules! Attempting to repair names..."
        )
        warnings.warn(
            f"{msg}, following entities {invalid_names} do not follow these rules! Attempting to repair names...",
            stacklevel=2,
        )

    if data_model_issues and stop_on_exception:
        msg = " ".join(data_model_issues)
        logging.error(msg)
        raise ValueError(msg)
    elif data_model_issues and not stop_on_exception:
        msg = " ".join(data_model_issues)
        msg += " Redefinitions will be skipped!"
        logging.warn(msg)
        warnings.warn(
            msg,
            stacklevel=2,
        )

    def _define_fields(property_definitions: list[Property]) -> dict[str, GraphQLField]:
        gql_field_definitions = {}
        for property_ in property_definitions:
            property_name = repair_name(property_.property_name, "property", fix_casing=fix_casing)  # type: ignore

            if property_name in gql_field_definitions:
                logging.warn(f"Property {property_name} being redefined... skipping!")
                warnings.warn(f"Property {property_name} being redefined... skipping!", stacklevel=2)
                continue

            # Node attribute
            if property_.property_type == "DatatypeProperty":
                value_type_gql = DATA_TYPE_MAPPING[property_.expected_value_type]["GraphQL"]

                # Case: Mandatory, single value
                if property_.min_count and property_.max_count == 1:
                    value = GraphQLNonNull(value_type_gql)
                # Case: Mandatory, multiple value
                elif property_.min_count and property_.max_count != 1:
                    value = GraphQLNonNull(GraphQLList(GraphQLNonNull(value_type_gql)))
                # Case: Optional, single value
                elif property_.max_count == 1:
                    value = value_type_gql
                # Case: Optional, multiple value
                else:
                    value = GraphQLList(value_type_gql)

            # Node edge
            else:
                value = gql_type_definitions[repair_name(property_.expected_value_type, "class", fix_casing=fix_casing)]
                is_one_to_many_edge = not (property_.min_count and property_.max_count == 1)
                if is_one_to_many_edge:
                    value = GraphQLList(value)
            gql_field_definitions[property_name] = GraphQLField(value)

        return gql_field_definitions

    gql_type_definitions: dict = {}
    for class_, properties in transformation_rules.get_classes_with_properties().items():
        gql_type_definitions[repair_name(class_, "class", fix_casing=fix_casing)] = GraphQLObjectType(
            repair_name(class_, "class", fix_casing=fix_casing),
            lambda properties=properties: _define_fields(properties),
        )

    # Needs this so we are able to generate the schema string
    query_schema = GraphQLSchema(
        query=GraphQLObjectType(
            "Query", lambda: {type_name: GraphQLField(type_def) for type_name, type_def in gql_type_definitions.items()}
        )
    )
    return _get_graphql_schema_string(query_schema)



from cognite.client import CogniteClient

from cognite.neat.core.workflow.model import FlowMessage


class WorkflowTaskBuilder:
    """Collection of all base tasks for workflows.All tasks must run in the context of a workflow including threads."""

    def __init__(self, cdf_client: CogniteClient, worflow_manager):
        # TODO : figure out ciclura import and set type to WorkflowManager
        self.cdf_client = cdf_client
        self.workflow_manager = worflow_manager

    def start_workflow_task(self, workflow_name: str, sync: bool, flow_message: FlowMessage) -> FlowMessage:
        """Call a workflow task from another workflow"""
        workflow = self.workflow_manager.get_workflow(workflow_name)
        return workflow.start(sync=sync, flow_message=flow_message)



from .base import BaseWorkflow, FlowMessage, WorkflowFullStateReport, WorkflowStepDefinition, WorkflowStepEvent
from .manager import WorkflowManager

__al__ = [
    "BaseWorkflow",
    "WorkflowStepDefinition",
    "WorkflowManager",
    "WorkflowStepEvent",
    "WorkflowFullStateReport",
    "FlowMessage",
]



import shutil
from pathlib import Path

from cognite.neat.constants import EXAMPLES_DIRECTORY


def copy_examples_to_directory(target_data_dir: Path):
    """
    Copier over all the examples to the target_data_directory,
    without overwriting


    Parameters
    ----------
    target_data_dir : The target directory


    """
    print(f"Copying examples into {target_data_dir}")
    for current in EXAMPLES_DIRECTORY.rglob("*"):
        if current.is_dir():
            continue
        relative = current.relative_to(EXAMPLES_DIRECTORY)
        if not (target := target_data_dir / relative).exists():
            target.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy2(current, target)



from .labels import upload_labels
from .rdf_to_assets import categorize_assets, rdf2assets, upload_assets
from .rdf_to_relationships import categorize_relationships, rdf2relationships, upload_relationships
from .rules_to_graphql import rules2graphql_schema

__all__ = [
    "rdf2relationships",
    "categorize_assets",
    "upload_assets",
    "rdf2assets",
    "categorize_relationships",
    "upload_relationships",
    "upload_labels",
    "rules2graphql_schema",
]
